﻿namespace Pikta.Editor.Tests
{
    using System;
    using System.Diagnostics.Contracts;
    using System.Globalization;
    using System.Linq;
    using Microsoft.VisualStudio.TestTools.UnitTesting;
    using Pikta.Expressions;

    [TestClass]
    public class CollectionAdapterTest
    {
        private Element collection;
        private CollectionAdapter adapter;

        [ClassInitialize]
        public static void ClassInitialize(TestContext context)
        {
            Contract.ContractFailed +=
                (s, e) =>
                {
                    e.SetUnwind();
                };
        }

        [TestInitialize]
        public void TestInitialize()
        {
            this.collection = Namespace.Global[KnownNames.Collection].Clone();
            this.adapter = new CollectionAdapter(this.collection);
        }

        [TestMethod]
        public void CollectionIndexOfMissing()
        {
            var e = Element.Base.Clone();
            var f = Element.Base.Clone();
            this.collection["20"].Value = new Constant(e);

            int index = this.adapter.IndexOf(f);

            Assert.AreEqual(-1, index);
        }

        [TestMethod]
        public void CollectionIndexOf()
        {
            var e = Element.Base.Clone();
            var f = Element.Base.Clone();
            this.collection["20"].Value = new Constant(e);
            this.collection["30"].Value = new Constant(f);

            int index = this.adapter.IndexOf(f);

            Assert.AreEqual(1, index);
        }

        [TestMethod]
        public void CollectionInsertToEmpty()
        {
            var e = Element.Base.Clone();

            this.adapter.Insert(0, e);

            Assert.AreSame(e, this.adapter.Single());
        }

        [TestMethod]
        public void CollectionInsertToStart()
        {
            var e = Element.Base.Clone();
            var f = Element.Base.Clone();

            this.adapter.Insert(0, e);
            this.adapter.Insert(0, f);

            var expected = new[] { f, e };
            var actual = this.adapter.ToArray();

            CollectionAssert.AreEqual(expected, actual);
        }

        [TestMethod]
        public void CollectionInsertToEnd()
        {
            var e = Element.Base.Clone();
            var f = Element.Base.Clone();

            this.adapter.Insert(0, e);
            this.adapter.Insert(1, f);

            var expected = new[] { e, f };
            var actual = this.adapter.ToArray();

            CollectionAssert.AreEqual(expected, actual);
        }

        [TestMethod]
        public void CollectionInsertOneSpace()
        {
            var e = Element.Base.Clone();
            var f = Element.Base.Clone();
            var g = Element.Base.Clone();
            this.collection["20"].Value = new Constant(e);
            this.collection["22"].Value = new Constant(f);

            this.adapter.Insert(1, g);

            var expected = new[] { e, g, f };
            var actual = this.adapter.ToArray();

            CollectionAssert.AreEqual(expected, actual);
        }

        [TestMethod]
        public void CollectionInsertNoSpaceInside()
        {
            var e = Element.Base.Clone();
            var f = Element.Base.Clone();
            var g = Element.Base.Clone();
            this.collection["20"].Value = new Constant(e);
            this.collection["21"].Value = new Constant(f);

            this.adapter.Insert(1, g);

            var expected = new[] { e, g, f };
            var actual = this.adapter.ToArray();

            CollectionAssert.AreEqual(expected, actual);
        }

        [TestMethod]
        public void CollectionInsertNoSpaceAtEnd()
        {
            var e = Element.Base.Clone();
            var f = Element.Base.Clone();
            var g = Element.Base.Clone();
            this.collection["0"].Value = new Constant(e);
            this.collection[int.MaxValue.ToString(CultureInfo.InvariantCulture)].Value = new Constant(f);

            this.adapter.Insert(2, g);

            var expected = new[] { e, f, g };
            var actual = this.adapter.ToArray();

            CollectionAssert.AreEqual(expected, actual);
        }

        [TestMethod]
        public void CollectionCount()
        {
            var e = Element.Base.Clone();
            var f = Element.Base.Clone();
            this.collection["20"].Value = new Constant(e);
            this.collection["30"].Value = new Constant(f);

            int actualCount = this.adapter.Count;

            Assert.AreEqual(2, actualCount);
        }

        [TestMethod]
        public void CollectionCopyTo()
        {
            var e = Element.Base.Clone();
            var f = Element.Base.Clone();
            this.collection["20"].Value = new Constant(e);
            this.collection["30"].Value = new Constant(f);

            var expected = new[] { e, f };

            var actual = new Element[3];
            this.adapter.CopyTo(actual, 1);
            actual = actual.Skip(1).ToArray();

            CollectionAssert.AreEqual(expected, actual);
        }

        [TestMethod]
        public void CollectionClear()
        {
            var e = Element.Base.Clone();
            var f = Element.Base.Clone();
            var g = Element.Base.Clone();
            this.adapter.Add(e);
            this.adapter.Add(f);
            this.adapter.Add(g);

            this.adapter.Clear();

            var expected = new Element[0];
            var actual = this.adapter.ToArray();

            CollectionAssert.AreEqual(expected, actual);
        }

        [TestMethod]
        public void CollectionEnumeration()
        {
            var e = Element.Base.Clone();
            var f = Element.Base.Clone();
            var g = Element.Base.Clone();
            this.collection["20"].Value = new Constant(e);
            this.collection["30"].Value = new Constant(f);
            this.collection["25"].Value = new Constant(g);

            var expected = new[] { e, g, f };
            var actual = this.adapter.ToArray();

            CollectionAssert.AreEqual(expected, actual);
        }

        [TestMethod]
        public void CollectionContains()
        {
            var e = Element.Base.Clone();
            var f = Element.Base.Clone();
            var g = Element.Base.Clone();
            this.collection["20"].Value = new Constant(e);
            this.collection["25"].Value = new Constant(g);

            Assert.IsTrue(this.adapter.Contains(g));
            Assert.IsFalse(this.adapter.Contains(f));
        }

        [TestMethod]
        public void CollectionRemoveAtStart()
        {
            var e = Element.Base.Clone();
            var f = Element.Base.Clone();
            var g = Element.Base.Clone();
            this.adapter.Add(e);
            this.adapter.Add(f);
            this.adapter.Add(g);

            this.adapter.RemoveAt(0);

            var expected = new[] { f, g };
            var actual = this.adapter.ToArray();

            CollectionAssert.AreEqual(expected, actual);
        }

        [TestMethod]
        public void CollectionRemoveAtEnd()
        {
            var e = Element.Base.Clone();
            var f = Element.Base.Clone();
            var g = Element.Base.Clone();
            this.adapter.Add(e);
            this.adapter.Add(f);
            this.adapter.Add(g);

            this.adapter.RemoveAt(2);

            var expected = new[] { e, f };
            var actual = this.adapter.ToArray();

            CollectionAssert.AreEqual(expected, actual);
        }

        [TestMethod]
        public void CollectionRemoveInside()
        {
            var e = Element.Base.Clone();
            var f = Element.Base.Clone();
            var g = Element.Base.Clone();
            this.adapter.Add(e);
            this.adapter.Add(f);
            this.adapter.Add(g);

            this.adapter.RemoveAt(1);

            var expected = new[] { e, g };
            var actual = this.adapter.ToArray();

            CollectionAssert.AreEqual(expected, actual);
        }

        [TestMethod]
        public void CollectionIndexerGet()
        {
            var e = Element.Base.Clone();
            var f = Element.Base.Clone();
            this.adapter.Add(e);
            this.adapter.Add(f);

            var expected = f;
            var actual = this.adapter[1];

            Assert.AreSame(expected, actual);
        }

        [TestMethod]
        public void CollectionIndexerSet()
        {
            var e = Element.Base.Clone();
            var f = Element.Base.Clone();
            var g = Element.Base.Clone();
            this.adapter.Add(e);
            this.adapter.Add(f);

            this.adapter[1] = g;

            var expected = new[] { e, g };
            var actual = this.adapter.ToArray();

            CollectionAssert.AreEqual(expected, actual);
        }
    }
}
