﻿namespace Pikta.Editor
{
    using System;
    using System.Collections.Generic;
    using System.Diagnostics.Contracts;
    using System.Drawing;
    using System.IO;
    using System.Linq;
    using System.Reactive.Linq;
    using System.Text;
    using System.Threading;

    /// <summary>
    /// Caches recently used images with automatic time-based cleaning.
    /// </summary>
    internal class BitmapCache
    {
        private readonly Dictionary<string, CacheEntry> cache = new Dictionary<string, CacheEntry>(StringComparer.OrdinalIgnoreCase);

        private IDisposable cleaningTimer;
        private string currentPath;

        public void SetCurrentPath(string currentPath)
        {
            string newPath = currentPath ?? string.Empty;

            if (this.currentPath != newPath)
            {
                this.currentPath = newPath;

                // clear the cache
                lock (this.cache)
                {
                    foreach (CacheEntry entry in this.cache.Values)
                    {
                        if (entry.Image != null)
                        {
                            entry.Image.Dispose();
                        }
                    }

                    this.cache.Clear();
                }
            }
        }

        public Image this[string fileName]
        {
            get
            {
                lock (this.cache)
                {
                    CacheEntry entry;

                    if (this.cache.TryGetValue(fileName, out entry))
                    {
                        entry.LastAccess = DateTime.Now;
                    }
                    else
                    {
                        entry = this.cache[fileName] =
                            new CacheEntry() { Creation = DateTime.Now, LastAccess = DateTime.Now, Image = TryLoadImage(fileName) };

                        if (this.cache.Count == 1)
                        {
                            StartCleaningTimer();
                        }
                    }

                    return entry.Image;
                }
            }
        }

        private void StartCleaningTimer()
        {
            Contract.Requires(this.cleaningTimer == null);
            this.cleaningTimer = Observable.Interval(TimeSpan.FromSeconds(10)).Subscribe(_ => Clean());
        }

        private void Clean()
        {
            if (!Monitor.TryEnter(this.cache))
            {
                return;
            }

            try
            {
                DateTime now = DateTime.Now;

                var entries = this.cache.ToArray();

                foreach (var pair in entries)
                {
                    // if the entry was not recently accessed,
                    if (now - pair.Value.LastAccess > TimeSpan.FromMinutes(1))
                    {
                        // remove it
                        if (pair.Value.Image != null)
                        {
                            pair.Value.Image.Dispose();
                        }
                        this.cache.Remove(pair.Key);
                    }
                    // else if the image was not successfuly loaded before a while,
                    else if(pair.Value.Image == null && now - pair.Value.Creation > TimeSpan.FromSeconds(10))
                    {
                        // try load it again
                        pair.Value.Image = TryLoadImage(pair.Key);
                        pair.Value.Creation = now;
                    }
                }

                // if the cache is empty, there is no need to run the timer
                if (this.cache.Count == 0)
                {
                    if (this.cleaningTimer != null)
                    {
                        this.cleaningTimer.Dispose();
                        this.cleaningTimer = null;
                    }
                }
            }
            finally
            {
                Monitor.Exit(this.cache);
            }
        }

        private Image TryLoadImage(string fileName)
        {
            try
            {
                return Image.FromFile(Path.Combine(currentPath, fileName));
            }
            catch (IOException) { return null; }
            catch (OutOfMemoryException) { return null; }
            catch (ArgumentException) { return null; }
            catch (UnauthorizedAccessException) { return null; }
        }

        private class CacheEntry
        {
            public Image Image { get; set; }

            public DateTime Creation { get; set; }

            public DateTime LastAccess { get; set; }
        }
    }
}
