﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Contracts;
using System.Globalization;
using System.Linq;
using System.Text;
using Pikta.Expressions;

namespace Pikta.Editor
{
    public class CollectionAdapter : IList<Element>
    {
        private readonly Element collection;

        public CollectionAdapter(Element collection)
        {
            this.collection = collection;
        }

        public int IndexOf(Element item)
        {
            int index = -1;

            foreach (Element e in this)
            {
                index++;

                if (e == item)
                {
                    return index;
                }
            }

            return -1;
        }

        public void Insert(int index, Element item)
        {
            int previousKey = GetPreviousKey(index);
            int nextKey = GetNextKey(index);

            // if there is a space between the two keys,
            if (nextKey - previousKey > 1)
            {
                int between = (int)(((long)nextKey + previousKey) / 2);

                if (between == previousKey)
                {
                    between++;
                }

                Contract.Assert(previousKey < between);
                Contract.Assert(between < nextKey);

                this.collection[between.ToString(CultureInfo.InvariantCulture)].Value = new Constant(item);
            }
            else
            {
                // otherwise rebuild the whole collection
                Rebuild(
                    this.GetItemProperties().Take(index).Select(p => p.Value)
                    .Concat(new[] { new Constant(item) })
                    .Concat(this.GetItemProperties().Skip(index).Select(p => p.Value)));
            }
        }

        private int GetNextKey(int index)
        {
            if (index < Count)
            {
                return GetItemKey(this.GetItemProperties().ElementAt(index));
            }
            else
            {
                return int.MaxValue;
            }
        }

        private int GetPreviousKey(int index)
        {
            if (index > 0)
            {
                return GetItemKey(this.GetItemProperties().ElementAt(index - 1));
            }
            else
            {
                return 0;
            }
        }

        public void RemoveAt(int index)
        {
            Property property = this.GetItemProperties().ElementAt(index);
            property.Remove();
        }

        public Element this[int index]
        {
            get
            {
                // The identity select prevents endless recursion, because without it the ElementAt method
                // would discover this indexer get and call it instead of enumerating.
                return this.Select(item => item).ElementAt(index);
            }
            set
            {
                Property property = this.GetItemProperties().ElementAt(index);
                property.Value = new Constant(value);
            }
        }

        public void Add(Element item)
        {
            Insert(Count, item);
        }

        public void Clear()
        {
            var allItems = this.collection.Properties.ToArray();

            foreach (Property p in allItems)
            {
                p.Remove();
            }
        }

        public bool Contains(Element item)
        {
            // The identity select prevents endless recursion, because without it the Enumerable.Contains method
            // would discover this Contains method and call it instead of enumerating.
            return this.Select(i => i).Contains(item);
        }

        public void CopyTo(Element[] array, int arrayIndex)
        {
            int index = arrayIndex;

            foreach (Element item in this)
            {
                array[index++] = item;
            }
        }

        public int Count
        {
            get { return this.collection.Properties.Count(); }
        }

        public bool IsReadOnly
        {
            get { return false; }
        }

        public bool Remove(Element item)
        {
            int index = IndexOf(item);

            if (index != -1)
            {
                RemoveAt(index);
                return true;
            }
            else
            {
                return false;
            }
        }

        public IEnumerator<Element> GetEnumerator()
        {
            return (from item in GetItemProperties()
                    let podelement = this.collection.GetElement(item.Name)
                    where podelement != null
                    select podelement).GetEnumerator();
        }

        private IEnumerable<Property> GetItemProperties()
        {
            return this.collection.Properties.OrderBy(GetItemKey);
        }

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        private int GetItemKey(Property item)
        {
            return GetItemKey(item.Name);
        }

        internal static int GetItemKey(string propertyName)
        {
            int index;

            if (int.TryParse(propertyName, NumberStyles.Integer, CultureInfo.InvariantCulture, out index))
            {
                return index;
            }
            else
            {
                return 0;
            }
        }

        private void Rebuild(IEnumerable<Expression> itemsExpressions)
        {
            if (!itemsExpressions.Any())
            {
                return;
            }

            // capture all elements
            var expressions = itemsExpressions.ToArray();

            // The interval length is not divided by .Length-1. This would cause
            // the last element to be placed at the end of the interval, thus every Add
            // would trigger rebuilding the collection.
            int interval = int.MaxValue / expressions.Length;

            // clear the list
            this.Clear();

            // re-add with new keys
            int key = 0;
            foreach (Expression item in expressions)
            {
                this.collection[key.ToString(CultureInfo.InvariantCulture)].Value = item;
                key += interval;
            }
        }
    }
}
