﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Contracts;
using System.Drawing;
using System.Linq;
using System.Text;
using Pikta.Expressions;

namespace Pikta.Editor
{
    public static class ElementExtensions
    {
        public static Color GetColor(this Element element, string propertyName, Color defaultValue = default(Color))
        {
            var barva = element.GetElement(propertyName);

            if (barva == null)
            {
                return defaultValue;
            }

            int a = GetIntNumber(barva, "a", 255);
            int r = GetIntNumber(barva, "r", 0);
            int g = GetIntNumber(barva, "g", 0);
            int b = GetIntNumber(barva, "b", 0);

            return Color.FromArgb(a, r, g, b);
        }

        public static Element ToElement(this Color color)
        {
            var e = Namespace.Global[KnownNames.Color].Clone();

            e["a"].Value = new Constant(color.A);
            e["r"].Value = new Constant(color.R);
            e["g"].Value = new Constant(color.G);
            e["b"].Value = new Constant(color.B);

            return e;
        }

        public static float GetFloatNumber(this Element e, string propertyName, float defaultValue = 1.0f)
        {
            return (float)GetNumber(e, propertyName, defaultValue);
        }

        public static int GetIntNumber(this Element e, string propertyName, int defaultValue = 1)
        {
            return (int)GetNumber(e, propertyName, defaultValue);
        }

        public static string GetString(this Element e, string propertyName, string defaultValue = "")
        {
            return GetValue(e, propertyName, defaultValue, result => Converters.ToTextOrNull(result) ?? defaultValue);
        }

        public static double GetNumber(this Element element, string propertyName, double defaultValue = 1)
        {
            return GetValue(element, propertyName, defaultValue, result => Converters.ToNumberOrNull(result) ?? defaultValue);
        }

        public static Element GetElement(this Element element, string propertyName, Element defaultValue = null)
        {
            return GetValue(element, propertyName, defaultValue, result => Converters.AsElement(result) ?? defaultValue);
        }

        [Pure]
        public static bool IsContainer(this Element element)
        {
            return element.IsDescendantOf(KnownNames.Container);
        }

        public static IList <Element> GetContainerChildren(this Element container)
        {
            Element obsah = container.GetElement(KnownNames.ContentProperty);

            if (obsah == null)
            {
                return new Element[0];
            }

            if (obsah.IsDescendantOf(KnownNames.Collection))
            {
                return new CollectionAdapter(obsah);
            }
            else
            {
                return new[] { obsah };
            }
        }

        private static T GetValue<T>(this Element element, string propertyName, T defaultValue, Func<Result, T> convertResult)
        {
            if (element.Defines(propertyName))
            {
                var result = element.Evaluate(propertyName);
                if (result.HasError)
                {
                    return defaultValue;
                }
                else
                {
                    return convertResult(result);
                }
            }
            else
            {
                return defaultValue;
            }
        }
    }
}
