﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Contracts;
using System.Linq;
using System.Text;

namespace Pikta.Editor
{
    /// <summary>
    /// Represents undoable operation.
    /// </summary>
    public class HistoryAction
    {
        /// <summary>
        /// Initializes the new instance of the <see cref="HistoryAction"/> class.
        /// </summary>
        /// <param name="name">The action descriptive name.</param>
        /// <param name="performAction">The action which performs (or redoes) the operation.</param>
        /// <param name="undoAction">The action which undoes the operation.</param>
        /// <param name="canRepeat">A value indicating whether the redo action can be executed repeatedly.</param>
        public HistoryAction(string name, Action performAction, Action undoAction, bool canRepeat = false)
        {
            Contract.Requires(name != null);
            Contract.Requires(performAction != null);
            Contract.Requires(undoAction != null);

            Name = name;
            PerformAction = performAction;
            UndoAction = undoAction;
            CanRepeat = canRepeat;
        }

        protected HistoryAction()
        {
        }

        /// <summary>
        /// Gets the action descriptive name.
        /// </summary>
        public string Name { get; protected set; }

        /// <summary>
        /// Gets a value indicating whether the action can be performed. If the value is <c>false</c>, the <see cref="DisableReason"/> property
        /// contains the reason description.
        /// </summary>
        public bool CanExecute { get { return string.IsNullOrEmpty(DisableReason); } }

        /// <summary>
        /// Gets the reason why the <see cref="CanExecute"/> property is <c>false</c>; otherwise null.
        /// </summary>
        public string DisableReason { get; protected set; }

        /// <summary>
        /// Gets the action which performs (or redoes) the operation.
        /// </summary>
        public Action PerformAction { get; protected set; }

        /// <summary>
        /// Gets the action which undoes the operation.
        /// </summary>
        public Action UndoAction { get; protected set; }

        /// <summary>
        /// Gets a value indicating whether the redo action can be executed repeatedly.
        /// </summary>
        public bool CanRepeat { get; protected set; }
    }
}
