﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Contracts;
using System.Linq;
using System.Reactive;
using System.Reactive.Subjects;
using System.Text;

namespace Pikta.Editor
{
    /// <summary>
    /// Manages history of performed actions.
    /// </summary>
    public class HistoryManager
    {
        private readonly Stack<HistoryAction> undoList = new Stack<HistoryAction>();
        private readonly Stack<HistoryAction> redoList = new Stack<HistoryAction>();
        private readonly Subject<Unit> changed = new Subject<Unit>();

        public bool CanUndo { get { return this.undoList.Count > 0; } }

        public bool CanRedo { get { return this.redoList.Count > 0; } }

        public bool CanRepeat { get { return this.redoList.Count == 0 && CanUndo && this.undoList.Peek().CanRepeat; } }

        public IEnumerable<string> UndoList { get { return this.undoList.Select(a => a.Name); } }

        public IEnumerable<string> RedoList { get { return this.redoList.Select(a => a.Name); } }

        public IObservable<Unit> Changed { get { return this.changed; } }

        public void Perform(HistoryAction action)
        {
            Contract.Requires(action != null);

            if (!action.CanExecute)
            {
                throw new InvalidOperationException(action.DisableReason);
            }

            action.PerformAction();

            this.undoList.Push(action);
            this.redoList.Clear();

            OnChanged();
        }

        public void Undo()
        {
            Contract.Requires(CanUndo);

            var undoedAction = this.undoList.Peek();

            undoedAction.UndoAction();

            this.undoList.Pop();
            this.redoList.Push(undoedAction);

            OnChanged();
        }

        public void Redo()
        {
            Contract.Requires(CanRedo);

            var redoedAction = this.redoList.Peek();
            
            redoedAction.PerformAction();

            this.redoList.Pop();
            this.undoList.Push(redoedAction);

            OnChanged();
        }

        public void Clear()
        {
            this.undoList.Clear();
            this.redoList.Clear();

            OnChanged();
        }

        private void OnChanged()
        {
            this.changed.OnNext(Unit.Default);
        }
    }
}
