﻿namespace Pikta.Editor
{
    using System;
    using System.Collections.Generic;
    using System.Diagnostics.Contracts;
    using System.Drawing;
    using System.Linq;
    using System.Reactive.Subjects;
    using System.Text;
    using System.Windows.Forms;

    public class InputProcessing : IDisposable
    {
        private readonly View view;
        private readonly Scene scene;
        private readonly Control outputControl;

        private readonly Subject<Element> selectClick = new Subject<Element>();
        private readonly Subject<Element> selectAddClick = new Subject<Element>();
        private readonly Subject<Element> elementContextMenuClick = new Subject<Element>();

        public InputProcessing(View view, Control outputControl)
        {
            this.view = view;
            this.scene = this.view.Scene;
            this.outputControl = outputControl;
            WireControl();
        }

        private DraggingHandler DraggingHandler { get; set; }

        private IList<Element> LastUnderCursor { get; set; }

        public bool IsDragging { get { return DraggingHandler != null; } }

        public IObservable<Element> SelectClick { get { return this.selectClick; } }

        public IObservable<Element> SelectAddClick { get { return this.selectAddClick; } }

        public IObservable<Element> ElementContextMenuClick { get { return this.elementContextMenuClick; } }

        public event EventHandler DraggingFinished;

        protected void OnDraggingFinished() { if (DraggingFinished != null) DraggingFinished(this, EventArgs.Empty); }

        public void OnExternalSelection()
        {
            LastUnderCursor = null;
        }

        public void Dispose()
        {
            UnwireControl();
        }

        private void WireControl()
        {
            outputControl.MouseDown += OutputControl_MouseDown;
            outputControl.MouseClick += OutputControl_MouseClick;
            outputControl.MouseDoubleClick += OutputControl_MouseDoubleClick;
            outputControl.MouseMove += OutputControl_MouseMove;
            outputControl.MouseUp += OutputControl_MouseUp;
        }

        private void UnwireControl()
        {
            outputControl.MouseDown -= OutputControl_MouseDown;
            outputControl.MouseClick -= OutputControl_MouseClick;
            outputControl.MouseDoubleClick -= OutputControl_MouseDoubleClick;
            outputControl.MouseMove -= OutputControl_MouseMove;
            outputControl.MouseUp -= OutputControl_MouseUp;
        }

        private void OutputControl_MouseDown(object sender, MouseEventArgs e)
        {
            SizeGripKind? clickedSizeGrip = this.view.Selection.SizeGripHitTest(e.Location);
            if (clickedSizeGrip != null)
            {
                // size grip click
                return;
            }

            Element clicked = GetSelectedElement(e);

            if (clicked != null && this.view.Selection.Selected.Contains(clicked))
            {
                // keep selection
            }
            else
            {
                if (IsSelectClick(e))
                {
                    this.selectClick.OnNext(clicked);
                }
                else if (IsSelectAddClick(e))
                {
                    this.selectAddClick.OnNext(clicked);
                }
            }
        }

        private Element GetSelectedElement(MouseEventArgs e)
        {
            var allUnderCursor = ElementHitTest(e.Location).ToArray();

            Element currentSelection = this.view.Selection.Selected.FirstOrDefault();

            Element clicked;

            if (currentSelection == null || !allUnderCursor.Contains(currentSelection) || (LastUnderCursor != null && !allUnderCursor.SequenceEqual(LastUnderCursor)))
            {
                // select new under cursor
                clicked = allUnderCursor.LastOrDefault();
            }
            else
            {
                clicked = currentSelection;

                //// select next element in allUnderCursor
                //int currentIndex = Array.LastIndexOf(allUnderCursor, currentSelection);
                //currentIndex--;
                //if (currentIndex == -1)
                //{
                //    currentIndex = allUnderCursor.Length - 1;
                //}
                //clicked = allUnderCursor[currentIndex];
            }

            LastUnderCursor = allUnderCursor;
            return clicked;
        }

        private void OutputControl_MouseMove(object sender, MouseEventArgs e)
        {
            if (e.Button != MouseButtons.None)
            {
                // neni-li dosud tazeni,
                if (DraggingHandler == null)
                {
                    // zkusime ho spustit
                    if (!TryStartDragging(e))
                        return;
                }

                DraggingHandler.Move(e);
            }
        }

        private bool TryStartDragging(MouseEventArgs e)
        {
            DraggingHandler handler;
            if ((handler = this.view.Selection.TryStartSizing(e)) != null)
            {
                DraggingHandler = handler;
            }
            else
            {
                // zjistime, zda ma nekdo zajem
                foreach (Element prvek in ElementHitTest(e.Location))
                {
                    if (HandleDraggingStart(prvek, e))
                        break;
                }
            }

            if (DraggingHandler != null)
            {
                DraggingHandler.Start(e);
                return true;
            }
            else
            {
                return false;
            }
        }

        private void OutputControl_MouseUp(object sender, MouseEventArgs e)
        {
            if (DraggingHandler != null)
            {
                DraggingHandler.Finish(e);
                DraggingHandler = null;

                OnDraggingFinished();
            }
            else
            {
                Element clicked = GetSelectedElement(e);

                if (IsSelectClick(e))
                {
                    this.selectClick.OnNext(clicked);
                }
                else if (IsSelectAddClick(e))
                {
                    this.selectAddClick.OnNext(clicked);
                }
            }
        }

        /// <summary>
        /// Returns all scene elements whose bounding rectangles contain specified point.
        /// </summary>
        /// <param name="sceneLocation">The point to hit test in scene coordinates.</param>
        public IEnumerable<Element> ElementHitTest(Point sceneLocation)
        {
            return (from e in this.scene.AllSceneElements
                    let r = this.view.GetElementBounds(e)
                    where r.Contains(sceneLocation)
                    select e);//.Distinct();
        }

        private void OutputControl_MouseClick(object sender, MouseEventArgs e)
        {
            if (!this.view.IsInteractive)
            {
                return;
            }

            // je-li to klik na konci tazeni, nebudem delat nic
            if (DraggingHandler != null) return;
        }

        private void OutputControl_MouseDoubleClick(object sender, MouseEventArgs e)
        {
        }

        private bool HandleDraggingStart(Element prvek, MouseEventArgs e)
        {
            Contract.Requires(DraggingHandler == null);
            Contract.Ensures(Contract.Result<bool>() ? DraggingHandler != null : true);

            //if (!Editor.CanInteractWith(prvek)) return false;

            var rodic = this.scene.GetParentContainer(prvek);
            if (rodic == null) return false;

            DraggingHandler = CreateContainerDragging(rodic, prvek);
            if (DraggingHandler == null)
                return false;

            //if(!HandleMouseClick(prvek, e))
            //{
            //    DraggingHandler = null;
            //    return false;
            //}

            return true;
        }

        private bool HandleMouseClick(Element element, MouseEventArgs e)
        {
            if (IsSelectClick(e))
            {
                this.selectClick.OnNext(element);
            }
            else if (IsSelectAddClick(e))
            {
                this.selectAddClick.OnNext(element);
            }
            else if (IsContextMenuClick(e))
            {
                //if (Editor.Selection.CanSelect(element))
                {
                    //if (!Editor.Selection.Selected.Contains(element))
                    //{
                    //    if (!Editor.SelectClick(element))
                    //        return false;
                    //}
                    this.elementContextMenuClick.OnNext(element);
                }
            }
            else
            {
                return false;
            }

            return true;
        }

        private static bool IsContextMenuClick(MouseEventArgs e)
        {
            return Control.ModifierKeys == Keys.None && e.Button == MouseButtons.Right;
        }

        private static bool IsSelectClick(MouseEventArgs e)
        {
            return Control.ModifierKeys == Keys.None && e.Button == MouseButtons.Left;
        }

        private static bool IsSelectAddClick(MouseEventArgs e)
        {
            return Control.ModifierKeys == Keys.Control && e.Button == MouseButtons.Left;
        }

        private DraggingHandler CreateContainerDragging(Element container, Element element)
        {
            if (container.IsDescendantOf(KnownNames.Canvas)) return new CanvasDraggingHandler(this.scene, this.view, element);

            return null;
        }
    }
}
