﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Pikta.Editor.Actions;
using Pikta.Editor.Definitions;
using Pikta.Expressions;

namespace Pikta.Editor
{
    /// <summary>
    /// Okno zobrazující editor scény.
    /// </summary>
    public partial class MainForm : Form, IActionTarget
    {
        private readonly PropertyEditor propertyEditor = new PropertyEditor();
        private readonly HistoryManager historyManager = new HistoryManager();
        private readonly string appVersion;
        
        private IDisposable historyChangedSubscription;
        private FileDocument document;

        public MainForm()
        {
            InitializeComponent();

            this.appVersion = GetAppVersionString();

            this.historyChangedSubscription = this.historyManager.Changed.Subscribe(_ => UpdateHistoryButtons());

            NewScene();
        }

        public Scene Scene { get; private set; }

        public View View { get; private set; }

        private Element CurrentElement { get { return this.elementContextMenuStrip.Tag as Element; } }

        private Property CurrentProperty { get { return this.propertyContextMenuStrip.Tag as Property; } }

        private string GetAppVersionString()
        {
            string exeName = System.Reflection.Assembly.GetEntryAssembly().Location;
            DateTime utcBuildTime = PEHelper.GetTimeStamp(exeName);

            return utcBuildTime.ToLocalTime().ToString("g");
        }

        private void InitializeScene()
        {
            Scene = new Scene(Namespace.Global);
            View = new View(this.sceneCanvas, Scene, this);
            View.IsInteractive = true;
            View.ElementSelected.Subscribe(SelectElementInTree);

            this.sceneTree.Scene = Scene;

            SceneFunctions.Define(this.Scene);
        }

        private void SelectElementInTree(Element element)
        {
            this.sceneTree.SelectedElement = element;
        }

        private bool CloseDocument()
        {
            if (this.document != null && this.document.NeedsSave)
            {
                bool? userChoice = Program.ShowYesNoCancel(string.Format("Chcete uložit změny provedené ve scéně {0}?", this.document.Title));

                if (userChoice.HasValue && userChoice.Value)
                {
                    if (!Save())
                    {
                        return false;
                    }
                }
                else if (userChoice == null)
                {
                    return false;
                }
            }

            // discard all previous elements
            Namespace.Global.Clear();

            this.historyManager.Clear();
            this.sceneTree.Scene = null;
            if (View != null)
            {
                View.Dispose();
                View = null;
            }
            this.document = null;

            UpdateWindowTitle();
            UpdateCommands();
            UpdateScene();

            return true;
        }

        private bool NewScene()
        {
            if (!CloseDocument())
            {
                return false;
            }

            LoadDefinitions(new ElementSerializer(Namespace.Global));

            Namespace.Global[KnownNames.Root] = CloneElement(Namespace.Global["Plátno"]);

            BitmapVisual.SetCurrentPath(null);

            InitializeScene();

            this.document = new FileDocument("bezjména");

            UpdateWindowTitle();
            UpdateCommands();
            UpdateScene();

            return true;
        }

        private bool Open()
        {
            if (!CloseDocument())
            {
                return false;
            }

            if (this.openFileDialog.ShowDialog(this) == DialogResult.OK)
            {
                return Open(this.openFileDialog.FileName);
            }
            else
            {
                return false;
            }
        }

        public bool Open(string fileName)
        {
            if (!CloseDocument())
            {
                return false;
            }

            try
            {
                // reload definitions
                var s = new ElementSerializer(Namespace.Global);
                LoadDefinitions(s);

                // load scene
                s.Load(fileName);
                BitmapVisual.SetCurrentPath(Path.GetDirectoryName(fileName));

                InitializeScene();

                this.document = new FileDocument(Path.GetFileNameWithoutExtension(fileName), fileName);

                UpdateWindowTitle();
                UpdateCommands();
                UpdateScene();
            }
            catch (Exception ex)
            {
                Program.ShowError(ex.Message);
                return false;
            }

            return true;
        }

        private void UpdateWindowTitle()
        {
            string appName = string.Format("{0} (verze {1})", Application.ProductName, this.appVersion);

            if (this.document != null)
            {
                this.Text = string.Format("{0}{1} - {2}",
                    this.document.Title,
                    this.document.NeedsSave ? "*" : string.Empty,
                    appName);
            }
            else
            {
                this.Text = appName;
            }
        }

        private bool Save(bool forceNaming = false)
        {
            string fileName = this.document.FileName;

            if (forceNaming || !this.document.IsNamed)
            {
                if (this.saveFileDialog.ShowDialog(this) == DialogResult.OK)
                {
                    fileName = this.saveFileDialog.FileName;
                }
                else
                {
                    return false;
                }
            }

            return Save(fileName);
        }

        private bool Save(string fileName)
        {
            try
            {
                new ElementSerializer(Namespace.Global).Save(this.Scene.Root, fileName);
                BitmapVisual.SetCurrentPath(Path.GetDirectoryName(fileName));
                this.document.Title = Path.GetFileNameWithoutExtension(fileName);
                this.document.SavedAs(fileName);
                UpdateWindowTitle();

                return true;
            }
            catch (Exception ex)
            {
                Program.ShowError("Soubor se nepodařilo uložit. " + ex.Message);
                return false;
            }
        }

        private static void LoadDefinitions(ElementSerializer s)
        {
            string exePath = Path.GetDirectoryName(System.Reflection.Assembly.GetEntryAssembly().Location);
            var definitions = new[]
                              {
                                  "Colors.pikta",
                                  "Shapes.pikta",
                                  "Containers.pikta",
                                  "Components.pikta"
                              }
                              .Select(file => Path.Combine(exePath, "Definitions", file));

            foreach (string def in definitions)
            {
                s.Load(def);
            }
        }

        private void editPropertyToolStripMenuItem_Click(object sender, EventArgs e)
        {
            EditProperty(CurrentProperty);
        }

        private void sceneTree_PropertyDoubleClick(object sender, PropertyEventArgs e)
        {
            EditProperty(e.Property);
        }

        private void EditProperty(Property property)
        {
            PerformAction(this.propertyEditor.EditExpression(property));
        }

        private void UpdateScene()
        {
            CheckSelectedElementsAvailability();
            this.sceneCanvas.Invalidate();
            this.sceneTree.UpdateTree();
        }

        private void CheckSelectedElementsAvailability()
        {
            if (this.Scene == null || this.View == null)
            {
                return;
            }

            var allElements = this.Scene.AllSceneElements.ToArray();
            var selected = this.View.Selection.Selected.ToArray();

            // remove selected elements not in the scene
            foreach (Element s in selected)
            {
                if (!allElements.Contains(s))
                {
                    this.View.Selection.Selected.Remove(s);
                }
            }
        }

        private void deleteElementToolStripMenuItem_Click(object sender, EventArgs e)
        {
            RemoveElement(CurrentElement);
        }

        private void RemoveElement(Element element)
        {
            // TODO: Najít všechny odkazy a opozornit, pokud na element něco odkazuje

            PerformAction(new RemoveElementAction(this.Scene, element));
        }

        public void PerformAction(HistoryAction action)
        {
            if (action == null)
            {
                return;
            }

            if (!action.CanExecute)
            {
                Program.ShowError(action.DisableReason, action.Name);
                return;
            }

            this.historyManager.Perform(action);
            RecordDocumentChange();
            UpdateScene();
        }

        private void RecordDocumentChange()
        {
            this.document.Changed();
            UpdateWindowTitle();
        }

        private void UpdateCommands()
        {
            saveToolStripButton.Enabled = saveToolStripMenuItem.Enabled =
                saveAsToolStripMenuItem.Enabled = exportToolStripMenuItem.Enabled = 
                printToolStripMenuItem.Enabled = this.document != null;

            generateGridToolStripMenuItem.Enabled = this.document != null;

            UpdateHistoryButtons();
        }

        private void UpdateHistoryButtons()
        {
            undoToolStripButton.Enabled = undoToolStripMenuItem.Enabled = this.historyManager.CanUndo;
            redoToolStripButton.Enabled = redoToolStripMenuItem.Enabled = this.historyManager.CanRedo;

            string undoCaption;

            if (this.historyManager.CanUndo)
            {
                undoCaption = string.Format("Zpět \"{0}\"", this.historyManager.UndoList.First());
            }
            else
            {
                undoCaption = "Zpět";
            }

            undoToolStripButton.Text = undoToolStripMenuItem.Text = undoCaption;

            string redoCaption;

            if (this.historyManager.CanRedo)
            {
                redoCaption = string.Format("Znovu \"{0}\"", this.historyManager.RedoList.First());
            }
            else
            {
                redoCaption = "Znovu";
            }

            redoToolStripButton.Text = redoToolStripMenuItem.Text = redoCaption;
        }

        private void undoToolStripButton_Click(object sender, EventArgs e)
        {
            this.historyManager.Undo();
            RecordDocumentChange();
            UpdateScene();
        }

        private void redoToolStripButton_Click(object sender, EventArgs e)
        {
            this.historyManager.Redo();
            RecordDocumentChange();
            UpdateScene();
        }

        private void removePropertyToolStripMenuItem_Click(object sender, EventArgs e)
        {
            RemoveProperty(CurrentProperty);
        }

        private void AddProperty(Element element)
        {
            PerformAction(this.propertyEditor.AddNewProperty(element));
        }

        private void RemoveProperty(Property property)
        {
            // TODO: Najít všechny odkazy a upozornit, pokud na vlastnost něco odkazuje a odstraní se poslední definice

            PerformAction(new RemovePropertyAction(property));
        }

        private void addPropertyToolStripMenuItem_Click(object sender, EventArgs e)
        {
            AddProperty(CurrentElement);
        }

        private void refreshButton_Click(object sender, EventArgs e)
        {
            UpdateScene();
        }

        private void elementContextMenuStrip_Opening(object sender, CancelEventArgs e)
        {
            addElementToolStripMenuItem.Enabled = CurrentElement.IsContainer();
            runGeneratorToolStripMenuItem.Enabled = CurrentElement.IsDescendantOf(KnownNames.Generator);
        }

        private void propertyContextMenuStrip_Opening(object sender, CancelEventArgs e)
        {
            editColorPropertyToolStripMenuItem.Enabled = this.propertyEditor.HasColorValue(CurrentProperty);
            removePropertyToolStripMenuItem.Enabled = CurrentProperty.IsLocal;
        }

        private void addElementToolStripMenuItem_Click(object sender, EventArgs e)
        {
            AddNewElement(CurrentElement);
        }

        private void cloneElementToolStripMenuItem_Click(object sender, EventArgs e)
        {
            var prototyp = CurrentElement;

            PerformAction(new AddElementAction(this.Scene, CloneElement(prototyp)));
        }

        private Element CloneElement(Element prototype)
        {
            var klon = prototype.Clone();

            if (prototype.IsContainer())
            {
                var novyObsah = Namespace.Global[KnownNames.Collection].Clone();
                var novaKolekce = new CollectionAdapter(novyObsah);
                klon[KnownNames.ContentProperty].Value = new Constant(novyObsah);

                foreach (Element subelement in prototype.GetContainerChildren())
                {
                    novaKolekce.Add(CloneElement(subelement));
                }
            }

            return klon;
        }

        private void AddNewElement(Element container)
        {
            using (var frm = new AddElementForm())
            {
                frm.AvailablePrototypes = this.Scene.AllElements.Where(e => e.GetName() != null);
                frm.Prototype = Element.Base;

                if (frm.ShowDialog(this) == DialogResult.OK)
                {
                    PerformAction(new AddElementAction(this.Scene, CloneElement(frm.Prototype), frm.ElementName, container));
                }
            }
        }

        private void renameToolStripMenuItem_Click(object sender, EventArgs e)
        {
            RenameElement(CurrentElement);
        }

        private void RenameElement(Element element)
        {
            string oldName = element.GetName();
            string newName = oldName;

            bool confirm = Program.ShowInput("Přejmenování elementu", "Zadejte nové jméno elementu:", ref newName);

            if (confirm && oldName != newName)
            {
                if (!string.IsNullOrEmpty(newName))
                {
                    if (!Parser.IsValidElementName(newName))
                    {
                        Program.ShowError("Zadaný název elementu není platný.");
                        return;
                    }
                }
                PerformAction(new RenameElementAction(this.Scene, element, newName.Trim()));
            }
        }

        private void sceneTree_ElementMoved(object sender, ElementMoveEventArgs e)
        {
            PerformAction(new MoveElementAction(this.Scene, e.SourceElement, e.TargetContainer));
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (!CloseDocument())
            {
                e.Cancel = true;
            }
        }

        private void newToolStripMenuItem_Click(object sender, EventArgs e)
        {
            NewScene();
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Open();
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Save();
        }

        private void saveAsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Save(forceNaming: true);
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void sceneTree_ElementMovedToPosition(object sender, ElementMoveToPositionEventArgs e)
        {
            PerformAction(new MoveElementToPositionAction(this.Scene, e.SourceElement, e.TargetPosition));
        }

        private void sceneAutoUpdateTimer_Tick(object sender, EventArgs e)
        {
            this.sceneCanvas.Invalidate();
        }

        private void sceneTree_PropertyMoved(object sender, PropertyMoveEventArgs e)
        {
            PerformAction(new SetPropertyReference(e.Target, e.Source)); 
        }

        private void runGeneratorToolStripMenuItem_Click(object sender, EventArgs e)
        {
            PerformAction(new RunGeneratorAction(this.Scene, this.sceneTree.SelectedElement, CloneElement));
        }

        private void editColorPropertyToolStripMenuItem_Click(object sender, EventArgs e)
        {
            PerformAction(this.propertyEditor.EditColor(this.sceneTree.SelectedProperty));
        }

        private void generateGridToolStripMenuItem_Click(object sender, EventArgs e)
        {
            GenerateGrid();
        }

        private void GenerateGrid()
        {
            // TODO: Co takhle ze dvou vybraných elementů vyplnit šablony?

            using (var frm = new GenerateGridForm())
            {
                frm.AvailableGridTemplates = this.Scene.AllContainers.Where(c => c.GetName() != null && (c.GetName() == KnownNames.Grid || c.IsDescendantOf(KnownNames.Grid)));
                frm.GridTemplate = Namespace.Global[KnownNames.Grid];

                frm.AvailableItemTemplates = this.Scene.AllElements.Where(t => t.GetName() != null).Except(frm.AvailableGridTemplates).OrderBy(t => t.GetName());
                frm.ItemTemplate = null;

                if (frm.ShowDialog(this) == DialogResult.OK)
                {
                    PerformAction(new GenerateGridAction(this.Scene, null, string.IsNullOrWhiteSpace(frm.GeneratedName) ? null : frm.GeneratedName,
                        frm.GridTemplate, frm.Rows, frm.Columns, frm.ItemTemplate, CloneElement));
                }
            }
        }

        private void periodicUpdateToolStripMenuItem_CheckedChanged(object sender, EventArgs e)
        {
            sceneAutoUpdateTimer.Enabled = periodicUpdateToolStripMenuItem.Checked;
        }

        private void sceneTree_ElementSelected(object sender, ElementActionEventArgs e)
        {
            this.View.Select(e.Element);
        }

        private void exportToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (this.exportImageFileDialog.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    Image image = this.View.DrawSceneToImage();
                    image.Save(this.exportImageFileDialog.FileName, ImageFormat.Png);
                }
                catch (Exception ex)
                {
                    Program.ShowError("Scénu se nepodařilo exportovat. " + ex.Message);
                }
            }
        }

        private void printToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (this.printDialog.ShowDialog() == DialogResult.OK)
            {
                printDocument.DocumentName = document.Title;
                printDocument.Print();
            }
        }

        private void printDocument_PrintPage(object sender, System.Drawing.Printing.PrintPageEventArgs e)
        {
            this.View.DrawSceneToPage(e.Graphics, e.MarginBounds);
        }
    }
}
