﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Forms;
using Ookii.Dialogs;
using Pikta.Expressions;

namespace Pikta.Editor
{
	static class Program
	{
		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main(string[] args)
		{
			Application.EnableVisualStyles();
			Application.SetCompatibleTextRenderingDefault(false);

            CommonFunctions.Define();

            var mainWindow = new MainForm();

            if (args.Length > 0)
            {
                mainWindow.Open(args[0]);
            }

			Application.Run(mainWindow);
		}

        public static void ShowError(string message, string title = null)
        {
            ShowMessageDialog(message, TaskDialogIcon.Error, title);
        }

        private static void ShowMessageDialog(string message, TaskDialogIcon icon = TaskDialogIcon.Custom, string title = null)
        {
            if (TaskDialog.OSSupportsTaskDialogs)
            {
                using (var dialog = new TaskDialog())
                {
                    dialog.Content = message;
                    dialog.WindowTitle = title ?? Application.ProductName;
                    dialog.MainIcon = icon;
                    dialog.CenterParent = true;
                    dialog.Buttons.Add(new TaskDialogButton(ButtonType.Close));
                    dialog.ShowDialog(ActiveForm);
                }
            }
            else
            {
                MessageBox.Show(ActiveForm, message, title ?? Application.ProductName, MessageBoxButtons.OK,
                    icon == TaskDialogIcon.Error ? MessageBoxIcon.Error : MessageBoxIcon.None);
            }
        }

        public static bool ConfirmWarning(string message)
        {
            if (TaskDialog.OSSupportsTaskDialogs)
            {
                using (var dialog = new TaskDialog())
                {
                    var yesButton = new TaskDialogButton(ButtonType.Yes);
                    var noButton = new TaskDialogButton(ButtonType.No) { Default = true };

                    dialog.Content = message;
                    dialog.WindowTitle = Application.ProductName;
                    dialog.MainIcon = TaskDialogIcon.Warning;
                    dialog.CenterParent = true;
                    dialog.Buttons.Add(yesButton);
                    dialog.Buttons.Add(noButton);
                    return dialog.ShowDialog(ActiveForm) == yesButton;
                }
            }
            else
            {
                return MessageBox.Show(ActiveForm, message, Application.ProductName, MessageBoxButtons.YesNo, MessageBoxIcon.Warning,
                    MessageBoxDefaultButton.Button2) == DialogResult.Yes;
            }
        }

        public static bool? ShowYesNoCancel(string message)
        {
            if (TaskDialog.OSSupportsTaskDialogs)
            {
                using (var dialog = new TaskDialog())
                {
                    var yesButton = new TaskDialogButton(ButtonType.Yes) { Default = true };
                    var noButton = new TaskDialogButton(ButtonType.No);
                    var cancelButton = new TaskDialogButton(ButtonType.Cancel);

                    dialog.Content = message;
                    dialog.WindowTitle = Application.ProductName;
                    dialog.MainIcon = TaskDialogIcon.Warning;
                    dialog.CenterParent = true;
                    dialog.Buttons.Add(yesButton);
                    dialog.Buttons.Add(noButton);
                    dialog.Buttons.Add(cancelButton);

                    var clickedButton = dialog.ShowDialog(ActiveForm);

                    if (clickedButton == yesButton)
                    {
                        return true;
                    }
                    else if (clickedButton == noButton)
                    {
                        return false;
                    }
                    else
                    {
                        return null;
                    }
                }
            }
            else
            {
                var ret = MessageBox.Show(ActiveForm, message, Application.ProductName, MessageBoxButtons.YesNoCancel, MessageBoxIcon.Warning);
                if (ret == DialogResult.Yes)
                {
                    return true;
                }
                else if (ret == DialogResult.No)
                {
                    return false;
                }
                else
                {
                    return null;
                }
            }
        }

        public static void ShowInfo(string message)
        {
            ShowMessageDialog(message);
        }

        public static bool ShowInput(string instruction, string message, ref string value)
        {
            using (var dialog = new InputDialog())
            {
                dialog.MainInstruction = instruction;
                dialog.Content = message;
                dialog.WindowTitle = Application.ProductName;
                dialog.Input = value;

                if (dialog.ShowDialog(ActiveForm) == DialogResult.OK)
                {
                    value = dialog.Input;
                    return true;
                }
                else
                {
                    return false;
                }
            }
        }

        private static Form ActiveForm
        {
            get
            {
                // to workaround issue https://connect.microsoft.com/VisualStudio/feedback/details/146673/unexpected-cross-thread-operation-not-valid-exception-when-running-in-the-debugger
                var form = Form.ActiveForm;
                if (form != null && Application.OpenForms.Cast<Form>().Contains(form))
                {
                    return form;
                }
                else
                {
                    return null;
                }
            }
        }
    }
}
