﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Contracts;
using System.Linq;
using System.Text;
using Pikta.Expressions;

namespace Pikta.Editor
{
	/// <summary>
	/// Scéna obsahující elementy.
	/// </summary>
	public class Scene
	{
		public Scene(Namespace sceneNamespace)
		{
			Namespace = sceneNamespace;

            if (Namespace.Contains(KnownNames.Root))
            {
                Root = Namespace[KnownNames.Root];
            }
            else
            {
                throw new InvalidOperationException("Scéna neobsahuje kořenový element.");
            }
		}

		public Namespace Namespace { get; private set; }

		public Element Root { get; private set; }

        /// <summary>
        /// Gets all elements reachable from the scene.
        /// </summary>
        public IEnumerable<Element> AllSceneElements
        {
            get { return EnumerateElementsBreadth(Root); }
        }

        /// <summary>
        /// Gets all container elements reachable from the scene.
        /// </summary>
        public IEnumerable<Element> AllSceneContainers
        {
            get { return AllSceneElements.Where(e => e.IsContainer()); }
        }

        /// <summary>
        /// Gets all elements reachable either from the scene or from the namespace.
        /// </summary>
        public IEnumerable<Element> AllElements
        {
            get { return AllSceneElements.Concat(Namespace.Global.Elements).SelectMany(e => new[] { e }.Concat(e.GetAncestors())).Distinct(); }
        }

        /// <summary>
        /// Gets all container elements reachable from the scene or from the namespace.
        /// </summary>
        public IEnumerable<Element> AllContainers
        {
            get { return AllElements.Where(e => e.IsContainer()); }
        }

        public Element GetParentContainer(Element element)
        {
            Contract.Requires(element != null);

            return (from c in AllSceneContainers
                    where c.GetContainerChildren().Contains(element)
                    select c).FirstOrDefault();
        }

        /// <summary>
        /// Returns all element that contain the specified element through the scene hierarchy.
        /// </summary>
        public IEnumerable<Element> GetSuperElementsOf(Element element)
        {
            var path = new Stack<Element>();
            LookupTree(path, Root, element);
            return path.Reverse();
        }

        /// <summary>
        /// Looks up the scene sub-tree for the specified element collecting path to it.
        /// </summary>
        private bool LookupTree(Stack<Element> path, Element root, Element element)
        {
            if (root == element)
            {
                return true;
            }

            path.Push(root);

            foreach (Element se in GetSubElements(root))
            {
                if (LookupTree(path, se, element))
                {
                    return true;
                }
            }

            path.Pop();

            return false;
        }

        /// <summary>
        /// Enumerates elements in the scene tree breadth-first.
        /// </summary>
        internal IEnumerable<Element> EnumerateElementsBreadth(Element parent)
        {
            Contract.Requires(parent != null);
            Contract.Ensures(Contract.Result<IEnumerable<Element>>() != null);

            Queue<Element> queue = new Queue<Element>();

            queue.Enqueue(parent);

            yield return parent;

            while (queue.Count > 0)
            {
                Element element = queue.Dequeue();

                foreach (Element child in GetSubElements(element))
                {
                    yield return child;

                    queue.Enqueue(child);
                }
            }
        }

        private IEnumerable<Element> GetSubElements(Element element)
        {
            return from def in element.LocalPropertyDefinitions.OrderBy(GetPropertyOrder)
                   from constant in def.Value.AllConstants
                   let constElement = constant.Value as Element
                   where constElement != null
                   select constElement;
        }

        private int GetPropertyOrder(PropertyDefinition definition)
        {
            return CollectionAdapter.GetItemKey(definition.Name);
        }

        /// <summary>
        /// Enumerates elements in the scene tree depth-first.
        /// </summary>
        private IEnumerable<Element> EnumerateElementsDepth(Element parent)
        {
            throw new NotImplementedException();

            //yield return parent;

            //if (parent.IsContainer())
            //{
            //    foreach (Element child in parent.GetContainerChildren())
            //    {
            //        foreach (Element item in EnumerateElementsDepth(child))
            //        {
            //            yield return item;
            //        }
            //    }
            //}
        }
    }
}
