﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Pikta.Expressions;
using System.Diagnostics.Contracts;

namespace Pikta.Editor
{
    public partial class SceneTree : UserControl
    {
        private Scene scene;
        private bool hideContainerContent = true;
        private ContextMenuStrip elementContextMenu;
        private ContextMenuStrip propertyContextMenu;

        public SceneTree()
        {
            InitializeComponent();
        }

        [Browsable(false)]
        public Scene Scene
        {
            get { return scene; }
            set
            {
                if (scene != value)
                {
                    scene = value;

                    UpdateTree();
                }
            }
        }

        [DefaultValue(true)]
        public bool HideContainerContent
        {
            get { return hideContainerContent; }
            set
            {
                if (hideContainerContent != value)
                {
                    hideContainerContent = value;
                    UpdateTree();
                }
            }
        }

        [DefaultValue(null)]
        public ContextMenuStrip ElementContextMenu
        {
            get { return elementContextMenu; }
            set
            {
                if (elementContextMenu != value)
                {
                    elementContextMenu = value;
                    UpdateTree();
                }
            }
        }

        [DefaultValue(null)]
        public ContextMenuStrip PropertyContextMenu
        {
            get { return propertyContextMenu; }
            set
            {
                if (propertyContextMenu != value)
                {
                    propertyContextMenu = value;
                    UpdateTree();
                }
            }
        }

        [Browsable(false)]
        public Element SelectedElement
        {
            get { return this.treeView.SelectedNode != null ? this.treeView.SelectedNode.Tag as Element : null; }
            set
            {
                this.treeView.SelectedNode = GetAllNodes(this.treeView.Nodes.Cast<TreeNode>()).FirstOrDefault(n => n.Tag == value);
            }
        }

        [Browsable(false)]
        public Property SelectedProperty
        {
            get { return this.treeView.SelectedNode != null ? this.treeView.SelectedNode.Tag as Property : null; }
        }

        [Category("Behavior")]
        public event EventHandler<PropertyEventArgs> PropertyDoubleClick;

        [Category("Behavior")]
        public event EventHandler<ElementMoveEventArgs> ElementMoved;

        [Category("Behavior")]
        public event EventHandler<PropertyMoveEventArgs> PropertyMoved;

        [Category("Behavior")]
        public event EventHandler<ElementMoveToPositionEventArgs> ElementMovedToPosition;

        [Category("Action")]
        public event EventHandler<ElementActionEventArgs> ElementSelected;

        public void UpdateTree()
        {
            treeView.BeginUpdate();

            TreeNode selectedNode = treeView.SelectedNode;

            treeView.Nodes.Clear();

            if (Scene != null && Scene.Root != null)
            {
                treeView.Nodes.Add(AddNode(Scene.Root, new List<Element>()));
            }

            // pokus o obnoveni vyberu
            if (selectedNode != null && selectedNode.Tag != null)
            {
                TreeNode nalezeny = GetAllNodes(treeView.Nodes.Cast<TreeNode>()).FirstOrDefault(n => n.Tag.Equals(selectedNode.Tag));
                if (nalezeny != null)
                {
                    nalezeny.EnsureVisible();
                    treeView.SelectedNode = nalezeny;

                    if (selectedNode.IsExpanded)
                    {
                        nalezeny.Expand();
                    }
                }
            }

            treeView.EndUpdate();

            UpdateButtons();
        }

        protected void OnPropertyDoubleClick(Property property)
        {
            Contract.Requires(property != null);

            var handler = PropertyDoubleClick;
            if (handler != null)
            {
                handler(this, new PropertyEventArgs(property));
            }
        }

        protected void OnElementMoved(Element sourceElement, Element targetContainer)
        {
            var handler = ElementMoved;
            if (handler != null)
            {
                handler(this, new ElementMoveEventArgs(sourceElement, targetContainer));
            }
        }

        protected void OnElementMovedToPosition(Element sourceElement, int targetPosition)
        {
            var handler = ElementMovedToPosition;

            if (handler != null)
            {
                handler(this, new ElementMoveToPositionEventArgs(sourceElement, targetPosition));
            }
        }

        protected void OnPropertyMoved(Property source, Property target)
        {
            var handler = PropertyMoved;
            if (handler != null)
            {
                handler(this, new PropertyMoveEventArgs(source, target));
            }
        }

        protected void OnElementSelected(Element element)
        {
            var handler = ElementSelected;
            if (handler != null)
            {
                handler(this, new ElementActionEventArgs(element));
            }
        }

        private void UpdateButtons()
        {
            Element container = this.SelectedElement != null ? this.Scene.GetParentContainer(this.SelectedElement) : null;

            if (container != null)
            {
                moveLeftToolStripButton.Enabled = this.Scene.GetParentContainer(container) != null;

                var children = container.GetContainerChildren();
                int currentPosition = children.IndexOf(this.SelectedElement);

                Element nextContainer = GetNextContainer(children, currentPosition);

                moveRightToolStripButton.Enabled = nextContainer != null;
                moveUpToolStripButton.Enabled = currentPosition > 0;
                moveDownToolStripButton.Enabled = currentPosition < children.Count - 1;
            }
            else
            {
                moveLeftToolStripButton.Enabled = moveRightToolStripButton.Enabled =
                    moveUpToolStripButton.Enabled = moveDownToolStripButton.Enabled = false;
            }
        }

        private Element GetNextContainer(IEnumerable<Element> children, int currentPosition)
        {
            return children.Skip(currentPosition + 1).Where(e => e.IsContainer()).FirstOrDefault();
        }

        private IEnumerable<TreeNode> GetAllNodes(IEnumerable<TreeNode> nodes)
        {
            foreach (TreeNode node in nodes)
            {
                yield return node;

                foreach (TreeNode child in GetAllNodes(node.Nodes.Cast<TreeNode>()))
                {
                    yield return child;
                }
            }
        }

        private TreeNode AddNode(Element element, ICollection<Element> addedElements, string nodeText = null)
        {
            var node = new TreeNode(nodeText ?? element.ToString());
            node.ImageKey = node.SelectedImageKey = GetElementImageKey(element);;
            node.Tag = element;
            node.Text = string.Format("{0} ({1})", node.Text, element.Prototype);

            // TODO: Toto způsobí zacyklení např. při vložení klonu kontejneru do jeho prototypu
            //// prototyp
            //if (element.Prototype != null)
            //{
            //    node.Nodes.Add(AddNode(element.Prototype, string.Format(": {0}", element.Prototype)));
            //}

            if (addedElements.Contains(element))
            {
                // continuing would introduce cycle
                return node;
            }

            addedElements.Add(element);

            // vlastnosti
            foreach (PropertyDefinition def in element.PropertyDefinitions.OrderBy(d => d.Name, NameComparer.Instance))
            {
                if (HideContainerContent)
                {
                    // obsah kontejneru preskocime, zobrazi se pak
                    if (element.IsDescendantOf(KnownNames.Container) && def.Name == KnownNames.ContentProperty)
                    {
                        continue;
                    }
                }

                Property vlastnost = new Property(element, def.Name);

                Expression hodnota = def.Value;

                // pridani uzlu vlastnosti
                var uzelVlastnosti = new TreeNode(string.Format("{0} = {1}", def.Name, hodnota));
                uzelVlastnosti.ImageKey = uzelVlastnosti.SelectedImageKey = "property";
                uzelVlastnosti.Tag = vlastnost;

                uzelVlastnosti.ForeColor = vlastnost.IsLocal ? SystemColors.WindowText : SystemColors.GrayText;

                node.Nodes.Add(uzelVlastnosti);

                var konstanta = hodnota as Constant;
                bool jeHodnotouElement = konstanta != null && konstanta.Value is Element;

                if (jeHodnotouElement)
                {
                    uzelVlastnosti.Nodes.Add(AddNode((Element)konstanta.Value, addedElements));
                }
            }

            // obsah kontejneru
            if (element.IsDescendantOf(KnownNames.Container))
            {
                foreach (Element podelement in new CollectionAdapter(Converters.ToElement(element.Evaluate(KnownNames.ContentProperty))))
                {
                    node.Nodes.Add(AddNode(podelement, addedElements));
                }
            }

            return node;
        }

        private string GetElementImageKey(Element element)
        {
            if (element.IsDescendantOf(KnownNames.Ellipse)) return "ellipse";
            else if (element.IsDescendantOf(KnownNames.Rectangle)) return "rectangle";
            else if (element.IsDescendantOf(KnownNames.Text)) return "text";
            else if (element.IsDescendantOf(KnownNames.Canvas)) return "canvas";
            else if (element.IsDescendantOf(KnownNames.Grid)) return "grid";
            else if (element.IsDescendantOf(KnownNames.Stack)) return "stack";
            else if (element.IsDescendantOf(KnownNames.Color)) return "color";
            else if (element.IsDescendantOf(KnownNames.Bitmap)) return "bitmap";
            else
            {
                return "element";
            }
        }

        private void treeView_NodeMouseClick(object sender, TreeNodeMouseClickEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                if (e.Node != null && e.Node.Tag is Element)
                {
                    OnElementSelected((Element)e.Node.Tag);
                }
            }
            else if (e.Button == MouseButtons.Right)
            {
                treeView.SelectedNode = e.Node;

                if (e.Node.Tag is Property && PropertyContextMenu != null)
                {
                    PropertyContextMenu.Tag = e.Node.Tag;
                    PropertyContextMenu.Show(this.treeView, e.Location);
                }
                else if (e.Node.Tag is Element && ElementContextMenu != null)
                {
                    ElementContextMenu.Tag = e.Node.Tag;
                    ElementContextMenu.Show(this.treeView, e.Location);
                }
            }
        }

        private void treeView_NodeMouseDoubleClick(object sender, TreeNodeMouseClickEventArgs e)
        {
            if (e.Node.Tag is Property)
            {
                OnPropertyDoubleClick((Property)e.Node.Tag);
            }
        }

        private void treeView_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.Modifiers == Keys.None)
            {
                if (e.KeyCode == Keys.Enter)
                {
                    if (this.SelectedProperty != null)
                    {
                        e.Handled = true;
                        OnPropertyDoubleClick(this.SelectedProperty);
                    }
                }
            }
        }

        private bool CanDrag(TreeNode sourceNode)
        {
            // for elements only scene container members can be drag
            var sourceElement = sourceNode.Tag as Element;

            if (sourceElement != null)
            {
                return this.Scene.GetParentContainer(sourceElement) != null;
            }

            var sourceProperty = sourceNode.Tag as Property;

            return sourceProperty != null;
        }

        private bool CanDragTo(TreeNode sourceNode, TreeNode targetNode)
        {
            if (!CanDrag(sourceNode))
            {
                return false;
            }

            if (sourceNode.Tag is Element)
            {
                var source = (Element)sourceNode.Tag;
                var target = targetNode.Tag as Element;

                return CanDragTo(source, target);
            }
            else if(sourceNode.Tag is Property)
            {
                var source = (Property)sourceNode.Tag;
                var target = targetNode.Tag as Property;

                return CanDragTo(source, target);
            }

            return false;
        }

        private bool CanDragTo(Property source, Property target)
        {
            if (source == target || target == null)
            {
                return false;
            }

            return true;
        }

        private bool CanDragTo(Element source, Element target)
        {
            if (target == null)
            {
                return false;
            }

            if (source == target)
            {
                return false;
            }

            bool isTargetSceneContainer = target.IsContainer() && (this.scene.Root == target || this.scene.GetParentContainer(target) != null);

            if (!isTargetSceneContainer)
            {
                // cannot move element to non-container element
                return false;
            }

            if (this.Scene.GetSuperElementsOf(target).Contains(source))
            {
                // cannot move parent to its child
                return false;
            }
            return true;
        }

        private void FinishDrop(TreeNode sourceNode, TreeNode targetNode)
        {
            if (sourceNode.Tag is Element)
            {
                var source = (Element)sourceNode.Tag;
                var target = (Element)targetNode.Tag;

                if (this.Scene.GetParentContainer(source) != target)
                {
                    OnElementMoved(source, target);
                }
            }
            else if (sourceNode.Tag is Property)
            {
                var source = (Property)sourceNode.Tag;
                var target = (Property)targetNode.Tag;

                OnPropertyMoved(source, target);
            }
        }

        private void treeView_ItemDrag(object sender, ItemDragEventArgs e)
        {
            // start the operation
            DoDragDrop(e.Item, DragDropEffects.Move | DragDropEffects.Scroll);
        }

        private void treeView_DragEnter(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(typeof(TreeNode)) && CanDrag((TreeNode)e.Data.GetData(typeof(TreeNode))))
            {
                e.Effect = e.AllowedEffect;
            }
            else
            {
                e.Effect = DragDropEffects.None;
            }
        }

        private void treeView_DragOver(object sender, DragEventArgs e)
        {
            if (!e.Data.GetDataPresent(typeof(TreeNode)))
            {
                return;
            }

            Point clientPoint = this.treeView.PointToClient(new Point(e.X, e.Y));
            TreeNode targetNode = this.treeView.GetNodeAt(clientPoint);

            if (targetNode != null && CanDragTo((TreeNode)e.Data.GetData(typeof(TreeNode)), targetNode))
            {
                //clientPoint.Offset(0, 30);
                //this.dragToolTip.Show("Popiska", this.treeView, clientPoint);
                this.treeView.SelectedNode = targetNode;
                e.Effect = e.AllowedEffect;
            }
            else
            {
                e.Effect = DragDropEffects.None;
            }

            // auto scroll
            if (clientPoint.Y + 10 > this.treeView.ClientSize.Height)
            {
                ScrollDown();
            }
            else if (clientPoint.Y < 10)
            {
                ScrollUp();
            }
        }

        private void ScrollUp()
        {
            SafeNativeMethods.SendMessage(this.treeView.Handle, SafeNativeMethods.WM_SCROLL, IntPtr.Zero, IntPtr.Zero);
        }

        private void ScrollDown()
        {
            SafeNativeMethods.SendMessage(this.treeView.Handle, SafeNativeMethods.WM_SCROLL, (IntPtr)1, IntPtr.Zero);
        }

        private void treeView_DragDrop(object sender, DragEventArgs e)
        {
            if (!e.Data.GetDataPresent(typeof(TreeNode)))
            {
                return;
            }

            var draggedNode = (TreeNode)e.Data.GetData(typeof(TreeNode));

            FinishDrop(draggedNode, this.treeView.SelectedNode);
        }

        private void treeView_AfterSelect(object sender, TreeViewEventArgs e)
        {
            UpdateButtons();
        }

        private void moveLeftToolStripButton_Click(object sender, EventArgs e)
        {
            Element parentContainer = this.Scene.GetParentContainer(this.SelectedElement);

            OnElementMoved(this.SelectedElement, this.Scene.GetParentContainer(parentContainer));
        }

        private void moveRightToolStripButton_Click(object sender, EventArgs e)
        {
            Element parentContainer = this.Scene.GetParentContainer(this.SelectedElement);
            var children = parentContainer.GetContainerChildren();
            Element nextContainer = GetNextContainer(children, children.IndexOf(this.SelectedElement));

            OnElementMoved(this.SelectedElement, nextContainer);
        }

        private void moveDownToolStripButton_Click(object sender, EventArgs e)
        {
            MoveSelectedElement(1);
        }

        private void moveUpToolStripButton_Click(object sender, EventArgs e)
        {
            MoveSelectedElement(-1);
        }

        private void MoveSelectedElement(int delta)
        {
            Element parentContainer = this.Scene.GetParentContainer(this.SelectedElement);
            int targetIndex = parentContainer.GetContainerChildren().IndexOf(this.SelectedElement) + delta;

            OnElementMovedToPosition(this.SelectedElement, targetIndex);
        }
    }
}
