﻿namespace Pikta.Editor
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Text;
    using System.Collections.ObjectModel;
    using System.Drawing;
    using System.Diagnostics.Contracts;
    using System.Drawing.Drawing2D;
    using System.Windows.Forms;

    /// <summary>
	/// Manages element selection layer.
	/// </summary>
	public class SelectionLayer : IViewLayer
	{
        private View view;

		public SelectionLayer(View view)
		{
			var vybrane = new ObservableCollection<Element>();
			vybrane.CollectionChanged += (s, e) => { View.Redraw(); OnSelectedChanged(); };
			Selected = vybrane;

			View = view;

			Grips = new List<Rectangle>();
		}

		public IList<Element> Selected { get; private set; }

		public View View
		{
			get { return view; }
			set
			{
                if (this.view != value)
                {
                    this.view = value;
                }
			}
		}

		public event EventHandler SelectedChanged;
		protected void OnSelectedChanged() { if(SelectedChanged != null) SelectedChanged(this, EventArgs.Empty); }

		public void Draw(Graphics g)
		{
			DrawSelected(g);
		}

		static Pen SelectionPen = new Pen(Color.Blue, 2) { DashStyle = DashStyle.Dash, LineJoin = LineJoin.Round };
        static Pen SelectionPenBack = new Pen(Color.White, 2) { LineJoin = LineJoin.Round };
        
		const int GripsPerElement = 3;

		/// <summary>Umístění úchytek pro změnu velikosti ve směru hodinových ručiček.</summary>
		private List<Rectangle> Grips { get; set; }

		private void DrawSelected(Graphics g)
		{
			Contract.Ensures(Grips.Count % GripsPerElement == 0);

			Grips.Clear();

			foreach(Element vybrany in Selected)
			{
                if (!vybrany.IsDescendantOf("Viditelný"))
                {
                    continue;
                }

				var r = View.GetElementBounds(vybrany);

				r.Inflate(5, 5);

                g.DrawRectangle(SelectionPenBack, r);
                g.DrawRectangle(SelectionPen, r);

                if (vybrany.IsDescendantOf(KnownNames.Shape))
                {
                    AddGrips(r);
                }
			}

            foreach (Rectangle gripRect in Grips)
            {
                DrawGrip(g, gripRect);
            }
		}

		private void DrawGrip(Graphics g, Rectangle gripRect)
		{
			g.FillRectangle(Brushes.White, gripRect);
			g.DrawRectangle(Pens.Blue, gripRect);
		}

		private void AddGrips(Rectangle er)
		{
			const int Size = 8;
			const int Offset = Size/2;

			// sirka
			Grips.Add(new Rectangle(er.Right - Offset, (er.Top + er.Bottom) / 2 - Offset, Size, Size));
			// sirka vyska
			Grips.Add(new Rectangle(er.Right - Offset, er.Bottom - Offset, Size, Size));
			// vyska
			Grips.Add(new Rectangle((er.Left + er.Right) / 2 - Offset, er.Bottom - Offset, Size, Size));
		}

		internal bool CanSelect(Element element)
		{
			return true;
		}

        public SizeGripKind? SizeGripHitTest(Point location)
        {
            for (int i = 0; i < Grips.Count; i++)
            {
                var r = Grips[i];
                r.Inflate(3, 3);
                if (r.Contains(location))
                    return GetKind(i);
            }
            return null;
        }

        internal DraggingHandler TryStartSizing(MouseEventArgs e)
        {
            var grip = SizeGripHitTest(e.Location);

            if (grip != null)
            {
                return new SizeDraggingHandler(View.Scene, View, grip.Value);
            }
            else
            {
                return null;
            }
        }

		SizeGripKind GetKind(int index)
		{
			switch(index % GripsPerElement)
			{
				case 0: return SizeGripKind.Width;
				case 1: return SizeGripKind.WidthHeight;
				case 2: return SizeGripKind.Height;
				default:
					throw new InvalidOperationException("Není zde přidán druh úchytky.");
			}
		}
	}
}
