﻿namespace Pikta.Editor
{
    using System;
    using System.Collections.Generic;
    using System.Diagnostics.Contracts;
    using System.Drawing;
    using System.Drawing.Drawing2D;
    using System.Linq;
    using System.Reactive.Subjects;
    using System.Text;
    using System.Windows.Forms;

    public class View : IVisualProvider, IDisposable
	{
        private readonly Control outputControl;
        private readonly Visual unknownVisual = new UnknownElementVisual();
        private readonly Subject<Element> elementSelected = new Subject<Element>();

        private readonly SelectionLayer selection;
        private readonly IList<Visual> handlers;
        private readonly InputProcessing input;

        private Point offset;

		public View(Control outputControl, Scene scene, IActionTarget actionTarget)
		{
            Contract.Requires(outputControl != null);
            Contract.Requires(scene != null);

            this.outputControl = outputControl;
            this.Scene = scene;
            this.ActionTarget = actionTarget;

            this.LastRectangles = new Dictionary<Element, Rectangle>();
            this.selection = new SelectionLayer(this);
            this.Layers = new List<IViewLayer>()
                {
                    this.selection
                };
            this.input = new InputProcessing(this, this.outputControl);

            // define visual handlers
            this.handlers = new Visual[]
                {
                    new Canvas(this),
                    new Stack(this),
                    new Grid(this),
                    new Box(),
                    new Ellipse(),
                    new Text(),
                    new BitmapVisual()
                };

            // wire handlers
            this.input.SelectClick.Subscribe(SelectClick);
            this.input.SelectAddClick.Subscribe(SelectAddClick);
            foreach (var h in this.handlers)
            {
                h.RectangleChanged.Subscribe(pair => UpdateLastRectangle(pair.Item1, pair.Item2));
            }

            WireOutputControl();
		}

		public Scene Scene { get; private set; }

        public IActionTarget ActionTarget { get; private set; }

        /// <summary>
        /// Gets or sets current mouse cursor over the view.
        /// </summary>
        public Cursor Cursor
        {
            get { return this.outputControl.Cursor; }
            set { this.outputControl.Cursor = value; }
        }

        public SelectionLayer Selection
        {
            get { return this.selection; }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the user can interact with the scene content.
        /// </summary>
        public bool IsInteractive { get; set; }

        /// <summary>
        /// Gets or sets the scene content offset relative to the output control.
        /// </summary>
        public Point Offset
        {
            get { return offset; }
            set
            {
                offset = value;
                Redraw();
            }
        }

        /// <summary>
        /// Occurs when an element is selected. In case of deselection the parameter is null.
        /// </summary>
        public IObservable<Element> ElementSelected { get { return this.elementSelected; } }

        public ICollection<IViewLayer> Layers { get; private set; }

        private IDictionary<Element, Rectangle> LastRectangles { get; set; }

        public Visual GetHandler(Element element)
        {
            var handler = handlers.FirstOrDefault(v => element.IsDescendantOf(v.Prototype));
            return handler ?? this.unknownVisual;
        }

        public Rectangle GetElementBounds(Element element)
        {
            if (element == null) return Rectangle.Empty;
            Rectangle r;
            LastRectangles.TryGetValue(element, out r);
            return r;
        }

        public void Select(Element element)
        {
            this.input.OnExternalSelection();
            this.SelectClick(element);
        }

        public void Redraw()
        {
            this.outputControl.Invalidate();
        }

        public void Dispose()
        {
            if (this.input != null)
            {
                this.input.Dispose();
            }
            this.elementSelected.OnCompleted();
            this.elementSelected.Dispose();
            UnwireOutputControl();
        }

        public Image DrawSceneToImage()
        {
            var rootDrawer = (Shape)GetHandler(Scene.Root);
            int width = rootDrawer.GetWidth(Scene.Root) + 1;
            int height = rootDrawer.GetHeight(Scene.Root) + 1;

            var bitmap = new Bitmap(width, height);
            using (var g = Graphics.FromImage(bitmap))
            {
                DrawSceneToPage(g, Rectangle.Empty);
            }

            return bitmap;
        }

        public void DrawSceneToPage(Graphics g, Rectangle bounds)
        {
            var rootDrawer = (Shape)GetHandler(Scene.Root);
            int width = rootDrawer.GetWidth(Scene.Root);
            int height = rootDrawer.GetHeight(Scene.Root);

            if (!bounds.IsEmpty)
            {
                g.TranslateTransform(
                    bounds.X + (bounds.Width - width) / 2,
                    bounds.Y + (bounds.Height - height) / 2);
            }

            g.SmoothingMode = SmoothingMode.HighQuality;
            rootDrawer.DrawElement(g, Scene.Root, Rectangle.Empty);
        }

        private void WireOutputControl()
        {
            this.outputControl.Paint += outputControl_Paint;
        }

        private void SelectClick(Element clicked)
        {
            this.selection.Selected.Clear();

            SelectAddClick(clicked);
        }

        private void SelectAddClick(Element clicked)
        {
            if (clicked != null)
            {
                if (!this.selection.Selected.Contains(clicked))
                {
                    this.selection.Selected.Add(clicked);
                }
            }

            this.elementSelected.OnNext(clicked);
        }

        private void UnwireOutputControl()
        {
            this.outputControl.Paint -= outputControl_Paint;
        }

        private void UpdateLastRectangle(Element element, Rectangle rectangle)
        {
            this.LastRectangles[element] = rectangle;
        }

		private void outputControl_Paint(object sender, PaintEventArgs e)
		{
			e.Graphics.SmoothingMode = SmoothingMode.HighQuality;

            e.Graphics.TranslateTransform(Offset.X, Offset.Y);

            // draw elements
            GetHandler(Scene.Root).DrawElement(e.Graphics, Scene.Root, Rectangle.Empty);

            // draw other layers
            foreach (var layer in this.Layers)
            {
                layer.Draw(e.Graphics);
            }
		}
    }
}
