﻿using System;
using System.Text;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Diagnostics.Contracts;
using Pikta.Expressions;

namespace Pikta.Tests
{
	[TestClass]
	public class EvaluationTest
	{
		[ClassInitialize]
		public static void Initialize(TestContext testContext)
		{
            CommonFunctions.Define();

			Contract.ContractFailed += Contract_ContractFailed;
		}

		private static void Contract_ContractFailed(object sender, ContractFailedEventArgs e)
		{
			Assert.Fail(e.Message);
			e.SetHandled();
		}

		[TestMethod]
		public void Constants()
		{
			var e = Element.Base;
			Result r;

			// cislo
			r = Evaluation.Evaluate(new Constant(12.0), e);
			Assert.IsFalse(r.HasError, r.ErrorMessage);
			Assert.AreEqual(r.Value, 12.0);
			// text
			r = Evaluation.Evaluate(new Constant("Ahoj"), e);
			Assert.IsFalse(r.HasError, r.ErrorMessage);
			Assert.AreEqual(r.Value, "Ahoj");
			// element
			r = Evaluation.Evaluate(new Constant(e), e);
			Assert.IsFalse(r.HasError, r.ErrorMessage);
			Assert.AreEqual(r.Value, e);
		}

		[TestMethod]
		public void BasicArithmetics()
		{
			var e = Element.Base;
			Result r;

			r = Evaluation.Evaluate(Parser.Parse("321 + 123"), e);
			Assert.IsFalse(r.HasError, r.ErrorMessage);
			Assert.AreEqual(444.0, r.Value);

			r = Evaluation.Evaluate(Parser.Parse("444 - 123"), e);
			Assert.IsFalse(r.HasError, r.ErrorMessage);
			Assert.AreEqual(321.0, r.Value);

			r = Evaluation.Evaluate(Parser.Parse("444 * 123"), e);
			Assert.IsFalse(r.HasError, r.ErrorMessage);
			Assert.AreEqual(444.0 * 123, r.Value);

			r = Evaluation.Evaluate(Parser.Parse("444 / 123"), e);
			Assert.IsFalse(r.HasError, r.ErrorMessage);
			Assert.AreEqual(444.0 / 123, r.Value);

			r = Evaluation.Evaluate(Parser.Parse("444 % 123"), e);
			Assert.IsFalse(r.HasError, r.ErrorMessage);
			Assert.AreEqual(444.0 % 123, r.Value);

			r = Evaluation.Evaluate(Parser.Parse("8 * (4 + 10)"), e);
			Assert.IsFalse(r.HasError, r.ErrorMessage);
			Assert.AreEqual(112.0, r.Value);

			r = Evaluation.Evaluate(Parser.Parse("8 * 4 + 10"), e);
			Assert.IsFalse(r.HasError, r.ErrorMessage);
			Assert.AreEqual(42.0, r.Value);

			r = Evaluation.Evaluate(Parser.Parse("10 + 8 * 4"), e);
			Assert.IsFalse(r.HasError, r.ErrorMessage);
			Assert.AreEqual(42.0, r.Value);
		}

		[TestMethod]
		public void BasicArithmeticErrors()
		{
			var e = Element.Base;
			Result r;

			r = Evaluation.Evaluate(Parser.Parse("15000 / 0"), e);
			Assert.IsTrue(r.HasError, "Division by zero succeeded resulting " + r.Value);

			r = Evaluation.Evaluate(Parser.Parse("15000 % 0"), e);
			Assert.IsTrue(r.HasError, "Modulo by zero succeeded resulting " + r.Value);

			r = Evaluation.Evaluate(new Multiplication(new Constant(double.MaxValue), new Constant(double.MaxValue)), e);
			Assert.IsTrue(r.HasError, "Max value {0} exceeded by {1}.", double.MaxValue, r.Value);
		}

        [TestMethod]
        public void CycleSelf()
        {
            var z = Element.Base.Clone();
            var e = z.Clone();
            z["vlastnost"].Value = new Constant(10);
            e["vlastnost"].Value = Parser.Parse(".vlastnost");

            Result r = e.Evaluate("vlastnost");
            
            Assert.IsTrue(r.HasError);
        }

        [TestMethod]
        public void CycleSelfComplex()
        {
            var z = Element.Base.Clone();
            var e = z.Clone();
            z["v2"].Value = new Constant(5);
            e["v2"].Value = Parser.Parse("(1 * (20 + .v2))");

            Result r = e.Evaluate("v2");
            
            Assert.IsTrue(r.HasError);
        }

        [TestMethod]
        public void CycleMutual()
        {
            var z = Element.Base.Clone();
            var e = z.Clone();
            z["c1"].Value = z["c2"].Value = new Constant(1);
            e["c1"].Value = Parser.Parse(".c2");
            e["c2"].Value = Parser.Parse(".c1");

            Result r = e.Evaluate("c1");
            Assert.IsTrue(r.HasError);
            
            r = e.Evaluate("c2");
            Assert.IsTrue(r.HasError);
        }

        [TestMethod]
        public void CycleFalse()
        {
            var z = Element.Base.Clone();
            z["cislo"].Value = new Constant(10);

            Result r = z.Evaluate(Parser.Parse(".cislo + .cislo"));

            Assert.IsFalse(r.HasError, "Double reference treated as cycle.");
        }

        [TestMethod]
        public void FunctionCall()
        {
            var z = Element.Base.Clone();

            double expected = 10 + Math.Sin(1);
            var r = Evaluation.Evaluate(Parser.Parse("10 + Sin(1)"), z);

            Assert.IsFalse(r.HasError);
            Assert.AreEqual(expected, r.Value);
        }

        [TestMethod]
        public void UnknownFunction()
        {
            var z = Element.Base.Clone();

            var r = Evaluation.Evaluate(Parser.Parse("2 * velkáNeznámá(19 * 30; 2)"), z);

            Assert.IsTrue(r.HasError);
        }
	}
}
