﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Pikta.Expressions;
using System.Globalization;
using System.Diagnostics.Contracts;
using System.Linq;

namespace Pikta.Tests
{
    /// <summary>
    ///This is a test class for ParserTest and is intended
    ///to contain all ParserTest Unit Tests
    ///</summary>
    [TestClass()]
    public class ParserTest
    {
        /// <summary>
        ///Gets or sets the test context which provides
        ///information about and functionality for the current test run.
        ///</summary>
        public TestContext TestContext { get; set; }

        private readonly CultureInfo cz = new CultureInfo("cs-CZ");

        #region Additional test attributes

        //Use ClassInitialize to run code before running the first test in the class
        [ClassInitialize]
        public static void MyClassInitialize(TestContext testContext)
        {
            Contract.ContractFailed += Contract_ContractFailed;
        }

        private static void Contract_ContractFailed(object sender, ContractFailedEventArgs e)
        {
            Assert.Fail(e.Message);
            e.SetHandled();
        }

        //Use ClassCleanup to run code after all tests in a class have run
        [ClassCleanup]
        public static void MyClassCleanup()
        {
            Contract.ContractFailed -= Contract_ContractFailed;
        }

        //
        //Use TestInitialize to run code before running each test
        //[TestInitialize()]
        //public void MyTestInitialize()
        //{
        //}
        //
        //Use TestCleanup to run code after each test has run
        //[TestCleanup()]
        //public void MyTestCleanup()
        //{
        //}
        //
        #endregion

        [TestMethod]
        public void IsValidSymbolTest()
        {
            // spravne vstupy
            Assert.IsTrue(Parser.IsValidSymbol("názevElementu"), "Nepřijme název elementu.");
            Assert.IsTrue(Parser.IsValidSymbol("Černá"), "Nepřijme název elementu.");
            Assert.IsTrue(Parser.IsValidSymbol(".názevVlastnosti"), "Nepřijme relativní název vlastnosti.");
            Assert.IsTrue(Parser.IsValidSymbol("názevElementu.názevVlastnosti"), "Nepřijme absolutní název vlastnosti.");
            Assert.IsTrue(Parser.IsValidSymbol("názevElementu.názevVlastnosti.názevVlastnosti.názevVlastnosti"), "Nepřijme vícenásobný absolutní název vlastnosti.");

            // spatne vstupy
            Assert.IsFalse(Parser.IsValidSymbol("	 "), "Neodmítne prázný text.");
            Assert.IsFalse(Parser.IsValidSymbol("nazev$Blbosti"), "Neodmítne symbol s neplatným znakem.");
            Assert.IsFalse(Parser.IsValidSymbol("nazevBlbosti."), "Neodmítne symbol s tečkou na konci.");
            Assert.IsFalse(Parser.IsValidSymbol("nazev.Blb osti"), "Neodmítne symbol s mezerou.");
        }

        [TestMethod]
        public void IsValidPropertyNameTest()
        {
            Assert.IsTrue(Parser.IsValidPropertyName("názevVlastnosti"));
            Assert.IsTrue(Parser.IsValidPropertyName("NázevElementu"));

            Assert.IsFalse(Parser.IsValidPropertyName(".názevVlastnosti"));
            Assert.IsFalse(Parser.IsValidPropertyName("něco.něcoJiného"));
            Assert.IsFalse(Parser.IsValidPropertyName("123"));
            Assert.IsFalse(Parser.IsValidPropertyName("něco a ještě něco"));
        }

        [TestMethod]
        public void IsValidElementNameTest()
        {
            Assert.IsTrue(Parser.IsValidPropertyName("názevVlastnosti"));
            Assert.IsTrue(Parser.IsValidPropertyName("NázevElementu"));

            Assert.IsFalse(Parser.IsValidPropertyName(".názevVlastnosti"));
            Assert.IsFalse(Parser.IsValidPropertyName("něco.něcoJiného"));
            Assert.IsFalse(Parser.IsValidPropertyName("123"));
            Assert.IsFalse(Parser.IsValidPropertyName("něco a ještě něco"));
        }

        [TestMethod]
        public void ParseTest()
        {
            // konstanty
            Assert.AreEqual(new Constant(123.0), Parser.Parse("123", cz), "Neparsuje celé číslo.");
            Assert.AreEqual(new Constant(123.43), Parser.Parse("123,43", cz), "Neparsuje desetinné číslo.");
            Assert.AreEqual(new Constant("text"), Parser.Parse("\"text\"", cz), "Neparsuje text.");
            // TODO: Element

            // základní aritmetika
            Assert.AreEqual(new Addition(new Constant(123), new Constant(210)), Parser.Parse("123 + 210", cz), "Neparsuje součet.");
            Assert.AreEqual(new Substraction(new Constant(123), new Constant(210)), Parser.Parse("123 - 210", cz), "Neparsuje rozdíl.");
            Assert.AreEqual(new Multiplication(new Constant(123), new Constant(210)), Parser.Parse("123 * 210", cz), "Neparsuje násobení.");
            Assert.AreEqual(new Division(new Constant(123), new Constant(210)), Parser.Parse("123 / 210", cz), "Neparsuje dělení.");
            Assert.AreEqual(new Modulo(new Constant(123), new Constant(210)), Parser.Parse("123 % 210", cz), "Neparsuje modulo.");

            // složené výrazy
            Assert.AreEqual(new Multiplication(new SymbolExpression("obdélník", "šířka"), new Addition(new Constant(20), new Constant(3500))),
                Parser.Parse("obdélník.šířka * (20 + 3500)"), "Neparsuje složený výraz.");
        }

        [TestMethod]
        public void Functions()
        {
            var sin0 = new FunctionDefinition("Sin", 0);
            var sin1 = new FunctionDefinition("Sin", 1);
            var sin2 = new FunctionDefinition("Sin", 2);

            // calls alone
            Assert.AreEqual(new FunctionCall(sin0, Enumerable.Empty<Expression>()), Parser.Parse("Sin()", cz));
            Assert.AreEqual(new FunctionCall(sin1, new[] { new Constant(1) }), Parser.Parse("Sin(1)", cz));
            Assert.AreEqual(new FunctionCall(sin2, new[] { new Constant(1), new Constant(2) }), Parser.Parse("Sin(1; 2)", cz));
            Assert.AreEqual(
                new FunctionCall(sin2, new[] { Parser.Parse("1+1", cz), Parser.Parse("4*(3+5)", cz) }),
                Parser.Parse("Sin(1+1; 4 * (3 + 5))", cz));

            var nestedCall =
                new FunctionCall(sin1,
                    new[]
                    {
                        new Addition(
                            new Constant(1),
                            new FunctionCall(sin2,
                                new[] { new Constant(2), new Constant(3) }))
                    });

            // nested call
            Assert.AreEqual(
                nestedCall,
                Parser.Parse("Sin(1 + Sin(2; 3))", cz));

            // complex call
            Assert.AreEqual(
                new Addition(
                    new SymbolExpression("Element", "vlastnost"),
                    new Multiplication(
                        nestedCall,
                        new Constant(2))),
                Parser.Parse("Element.vlastnost + Sin(1 + Sin(2; 3)) * 2", cz));
        }

        [TestMethod]
        [ExpectedException(typeof(ParsingException))]
        public void FunctionNoParams()
        {
            Parser.Parse("1 + Funkce(;)", cz);
        }

        [TestMethod]
        [ExpectedException(typeof(ParsingException))]
        public void SingleSemicolon()
        {
            Parser.Parse("2 * Sin 1 ; 1", cz);
        }

        [TestMethod]
        [ExpectedException(typeof(ParsingException))]
        public void SingleRightParenthesis()
        {
            Parser.Parse("Neco.nepovedeneho )", cz);
        }
    }
}
