﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics.Contracts;

namespace Pikta
{
	/// <summary>
	/// Objekt protypového systému Pikta.
	/// </summary>
	public class Element
	{
		private static Element baseElement = new Element();

		private Element()
		{
			LocalPropertyDefinitions = new List<PropertyDefinition>();
		}

		private Element(Element prototype)
			: this()
		{
			Contract.Requires(prototype != null);

			Prototype = prototype;
		}

		/// <summary>
		/// Vrátí základní element prototypového systému Pikta, který je předkem všech elementů (kromě sebe sama).
		/// </summary>
		public static Element Base { get { return baseElement; } }

		/// <summary>
		/// Vrátí prototyp tohoto elementu.
		/// </summary>
		public Element Prototype { get; private set; }

		/// <summary>
		/// Vrátí kolekci definic vlastností, které tento element definuje.
		/// </summary>
		public ICollection<PropertyDefinition> LocalPropertyDefinitions { get; private set; }

		/// <summary>
		/// Vrátí kolekci definic vlastností, které tento element definuje nebo je zdědil.
		/// </summary>
		public IEnumerable<PropertyDefinition> PropertyDefinitions
		{
			get
			{
				if(Prototype != null)
				{
					return LocalPropertyDefinitions.Union(from pd in Prototype.PropertyDefinitions
														  where LocalPropertyDefinitions.All(local => !NameComparer.Instance.Equals(local.Name, pd.Name))
														  select pd);
				}
				else
				{
					return LocalPropertyDefinitions;
				}
			}
		}

		/// <summary>
		/// Vrátí seznam vlastností, které tento element definuje nebo je zdědil.
		/// </summary>
		public IEnumerable<Property> Properties
		{
			get
			{
				return from pd in PropertyDefinitions
					   select new Property(this, pd.Name);
			}
		}

		/// <summary>
		/// Vrátí vlastnost tohoto elementu se zadaným názvem.
		/// </summary>
		/// <param name="propertyName">Název vlastnosti, která se má vrátit.</param>
		public Property this[string propertyName]
		{
			get
			{
				Contract.Requires(!string.IsNullOrEmpty(propertyName));
				var property = Properties.FirstOrDefault(p => NameComparer.Instance.Equals(p.Name, propertyName));
				if(property != null)
				{
					return property;
				}
				else
				{
					// vratime nedefinovanou vlastnost
					return new Property(this, propertyName);
				}
			}
		}

		/// <summary>
		/// Vrátí hodnotu určující, zda element nebo jeho předek definuje vlastnost se zadaným názvem.
		/// </summary>
		/// <param name="propertyName">Název vlastnosti.</param>
		public bool Defines(string propertyName)
		{
			return PropertyDefinitions.Any(pd => NameComparer.Instance.Equals(pd.Name, propertyName));
		}

        /// <summary>
        /// Vrátí hodnotu určující, zda element definuje vlastnost se zadaným názvem.
        /// </summary>
        /// <param name="propertyName">Název vlastnosti.</param>
        public bool DefinesDirectly(string propertyName)
        {
            return LocalPropertyDefinitions.Any(pd => NameComparer.Instance.Equals(pd.Name, propertyName));
        }

		/// <summary>
		/// Pokud element definuje vlastnost se zadaným názvem, naplní ji do parametru <paramref name="property"/> a vrátí true.
		/// V opačném případě vrátí false.
		/// </summary>
		/// <param name="name">Název vlastnosti, která se má vrátit.</param>
		/// <param name="property">Obdrží vlastnost se zadaným názvem, pokud ji element definuje. Jinak obdrží null.</param>
		/// <returns>Hodnota určující, zda element vlastnost definuje.</returns>
		public bool TryGetProperty(string name, out Property property)
		{
			property = Properties.FirstOrDefault(p => NameComparer.Instance.Equals(p.Name, name));
			return property != null;
		}

		/// <summary>
		/// Vrátí nový element, jehož prototypem je tento element.
		/// </summary>
		public Element Clone()
		{
			Contract.Ensures(Contract.Result<Element>() != null);
			return new Element(this);
		}

		public override string ToString()
		{
			// odkaz na namespace sem nepatri; lec je to nejjednodussi
			return Namespace.Global.GetName(this) ?? "<anonymní>";
		}

		/// <summary>
		/// Vrátí definici vlastnosti se zadaným názvem. Pokud není definována, vrátí null.
		/// </summary>
		internal PropertyDefinition GetPropertyDefinition(string propertyName)
		{
			Element element = this;

			while(element != null)
			{
				var definice = element.LocalPropertyDefinitions.SingleOrDefault(pd => NameComparer.Instance.Equals(pd.Name, propertyName));

				if(definice != null)
				{
					return definice;
				}
				else
				{
					element = element.Prototype;
				}
			}

			return null;
		}
	}
}
