﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Linq;
using Pikta.Expressions;
using System.Globalization;
using System.Diagnostics.Contracts;

namespace Pikta
{
    /// <summary>
    /// Ukládá a načítá elementy obsažené ve jmenném prostoru.
    /// </summary>
    public class ElementSerializer
    {
        public ElementSerializer(Namespace targetNamespace)
        {
            Namespace = targetNamespace;
        }

        public Namespace Namespace { get; private set; }

        public void Load(string fileName)
        {
            var doc = XDocument.Load(fileName);

            var readElements = new Dictionary<int, ElementInfo>();

            // load element trees
            ICollection<ElementInfo> rootInfos = doc.Root.Elements("Element").Select(xe => ToElementInfo(xe, readElements)).ToArray();

            // get flattened list of all elements
            ICollection<ElementInfo> allInfos = rootInfos.SelectMany(CollectAllInfos).Distinct().ToArray();

            // materialize all elements
            Materialize(allInfos);
        }

        public void Save(Element element, string fileName)
        {
            int lastId = 0;

            var doc = new XDocument(
                new XElement("PiktaDocument",
                    FromElement(element, ref lastId, new Dictionary<Element, int>())));

            doc.Save(fileName);
        }

        private XElement FromElement(Element element, ref int lastId, IDictionary<Element, int> writtenElements)
        {
            var xe = new XElement("Element");

            xe.Add(new XAttribute("id", ++lastId));

            writtenElements.Add(element, lastId);

            if (element.GetName() != null)
            {
                xe.Add(new XAttribute("name", element.GetName()));
            }

            if (element.Prototype != Element.Base)
            {
                string prototypeName = element.Prototype.GetName();

                // if the prototype has name,
                if(prototypeName != null)
                {
                    // write attribute with name
                    xe.Add(new XAttribute("prototype", prototypeName));
                }
                // else if if has been written,
                else if(writtenElements.ContainsKey(element.Prototype))
                {
                    // write attribute with id
                    xe.Add(new XAttribute("prototype", writtenElements[element.Prototype]));
                }
                else
                {
                    // write element tagged with new id
                    xe.Add(new XElement("Prototype", FromElement(element.Prototype, ref lastId, writtenElements)));
                }
            }

            foreach (var def in element.LocalPropertyDefinitions)
            {
                object xvalue;
                Element constantElement = def.Value is Constant ? ((Constant)def.Value).Value as Element : null;

                if (constantElement != null)
                {
                    if (writtenElements.ContainsKey(constantElement))
                    {
                        xvalue = new XElement("ElementRef", new XAttribute("id", writtenElements[constantElement]));
                    }
                    else
                    {
                        xvalue = FromElement(constantElement, ref lastId, writtenElements);
                    }
                }
                else
                {
                    xvalue = new XAttribute("value", def.Value.ToString(CultureInfo.InvariantCulture));
                }

                xe.Add(new XElement("Property", new XAttribute("name", def.Name), xvalue));
            }

            return xe;
        }

        private ElementInfo ToElementInfo(XElement element, IDictionary<int, ElementInfo> readElements)
        {
            var ei = new ElementInfo();

            var xid = element.Attribute("id");
            if (xid != null)
            {
                readElements.Add((int)xid, ei);
            }

            ei.Name = (string)element.Attribute("name");

            var xprototype = element.Attribute("prototype");
            // if it has prototype attribute,
            if(xprototype != null)
            {
                // get prototype info by name or id
                int prototypeId;

                if (int.TryParse(xprototype.Value, NumberStyles.Integer, CultureInfo.InvariantCulture, out prototypeId))
                {
                    ei.PrototypeInfo = readElements[prototypeId];
                }
                else
                {
                    ei.PrototypeName = xprototype.Value;
                }
            }
            // else if it has prototype element,
            else if(element.Element("Prototype") != null)
            {
                // read prototype info
                var xp = element.Element("Prototype");
                ei.PrototypeInfo = ToElementInfo(xp, readElements);
            }
            else
            {
                ei.PrototypeName = Element.Base.GetName();
            }

            foreach (XElement xprop in element.Elements("Property"))
            {
                string nazevVlastnosti = xprop.Attribute("name").Value;

                if (xprop.Attribute("value") != null)
                {
                    ei.Properties.Add(new PropertyInfo(nazevVlastnosti, Parser.Parse((string)xprop.Attribute("value"), CultureInfo.InvariantCulture)));
                }
                else
                {
                    if (xprop.Element("ElementRef") != null)
                    {
                        int refId = (int)xprop.Element("ElementRef").Attribute("id");
                        ei.Properties.Add(new PropertyInfo(nazevVlastnosti, readElements[refId]));
                    }
                    else
                    {
                        ei.Properties.Add(new PropertyInfo(nazevVlastnosti, ToElementInfo(xprop.Element("Element"), readElements)));
                    }
                }
            }

            return ei;
        }

        private IEnumerable<ElementInfo> CollectAllInfos(ElementInfo root)
        {
            yield return root;

            foreach (var pi in root.Properties)
            {
                if (pi.Constant != null)
                {
                    foreach (var child in CollectAllInfos(pi.Constant))
                        yield return child;
                }
            }
        }

        private void Materialize(ICollection<ElementInfo> allElements)
        {
            foreach (ElementInfo ei in allElements)
            {
                Materialize(allElements, ei);
            }
        }

        private Element Materialize(ICollection<ElementInfo> allElements, ElementInfo elementInfo)
        {
            // if its not materialized yet,
            if (elementInfo.Element == null)
            {
                Element prototype;

                // find prototype info
                ElementInfo prototypeInfo = elementInfo.PrototypeInfo ?? allElements.FirstOrDefault(e => NameComparer.Instance.Equals(e.Name, elementInfo.PrototypeName));
                if (prototypeInfo != null)
                {
                    // materialize prototype
                    prototype = Materialize(allElements, prototypeInfo);
                }
                else
                {
                    // get prototype directly from the namespace
                    if (!Namespace.TryGetElement(elementInfo.PrototypeName, out prototype))
                    {
                        throw new InvalidOperationException(string.Format("Prototyp \"{0}\" elementu \"{1}\" nebyl definován.", elementInfo.PrototypeName, elementInfo.Name));
                    }
                }

                // clone
                Element element = prototype.Clone();

                // name the element
                if (elementInfo.Name != null)
                {
                    Namespace[elementInfo.Name] = element;
                }

                elementInfo.Element = element;

                // fill in properties
                foreach (PropertyInfo pi in elementInfo.Properties)
                {
                    Expression value = pi.Expression ?? new Constant(Materialize(allElements, pi.Constant));

                    element[pi.Name].Value = value;
                }
            }

            return elementInfo.Element;
        }

        private class PropertyInfo
        {
            public PropertyInfo(string name, Expression expression)
            {
                this.Name = name;
                this.Expression = expression;
            }

            public PropertyInfo(string name, ElementInfo constant)
            {
                this.Name = name;
                this.Constant = constant;
            }

            public string Name { get; private set; }
            public Expression Expression { get; private set; }
            public ElementInfo Constant { get; private set; }
        }

        private class ElementInfo
        {
            public ElementInfo()
            {
                Properties = new List<PropertyInfo>();
            }

            public string Name { get; set; }
            public ElementInfo PrototypeInfo { get; set; }
            public string PrototypeName { get; set; }

            public IList<PropertyInfo> Properties { get; private set; }

            public Element Element { get; set; }
        }
    }
}
