﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Contracts;
using System.Linq;
using System.Text;

namespace Pikta.Expressions
{
    public class Evaluation
    {
        public delegate Result FunctionBody(Element context, IEnumerable<Result> args);

        static Evaluation()
        {
            FunctionsLibrary = new Dictionary<FunctionDefinition, FunctionBody>();
        }

        public Evaluation(Expression expression, Element context)
            : this(expression, context, null)
        {
            Contract.Requires(expression != null);
            Contract.Requires(context != null);
        }

        private Evaluation(Expression expression, Element context, IEnumerable<Expression> seenExpressions)
        {
            this.RootExpression = expression;
            this.Context = context;
            this.SeenExpressions = new HashSet<Expression>(
                seenExpressions ?? Enumerable.Empty<Expression>(),
                ReferenceEqualityComparer<Expression>.Instance);
        }

        public Element Context { get; private set; }
        public Expression RootExpression { get; private set; }

        public static IDictionary<FunctionDefinition, FunctionBody> FunctionsLibrary { get; private set; }

        private ISet<Expression> SeenExpressions { get; set; }

        /// <summary>
        /// Vyhodnotí výraz a vrátí jeho hodnotu.
        /// </summary>
        /// <param name="expression">Výraz, který se má vyhodnotit.</param>
        /// <param name="context">Element, v jehož kontextu se má výraz vyhodnotit.</param>
        public static Result Evaluate(Expression expression, Element context)
        {
            return new Evaluation(expression, context).Evaluate(expression);
        }

        private static Result Evaluate(Expression expression, Element context, IEnumerable<Expression> seenExpressions)
        {
            var eval = new Evaluation(expression, context, seenExpressions);
            return eval.Evaluate(expression);
        }

        private Result Evaluate(Expression expression)
        {
            var constant = expression as Constant;
            var unary = expression as UnaryExpression;
            var symbol = expression as SymbolExpression;
            var binary = expression as BinaryExpression;
            var function = expression as FunctionCall;

            if (constant != null)
            {
                return EvaluateConstant(constant);
            }
            else
            {
                if (SeenExpressions.Contains(expression))
                {
                    return Result.Error(string.Format("Byl zjištěn cyklus. Výraz {0} byl vyhodnocován podruhé.", expression));
                }

                SeenExpressions.Add(expression);

                if (unary != null)
                {
                    return EvaluateUnary(unary);
                }
                else if (symbol != null)
                {
                    return EvaluateSymbol(symbol);
                }
                else if (binary != null)
                {
                    return EvaluateBinary(binary);
                }
                else if (function != null)
                {
                    return EvaluateFunction(function);
                }
                else
                {
                    throw GetUnknownExpressionException();
                }
            }
        }

        private Result EvaluateFunction(FunctionCall function)
        {
            if (function.Arguments.Count != function.Definition.Arity)
            {
                return Result.Error(string.Format("Funkce {0} očekává {1} parametrů.", function.Definition.Name, function.Definition.Arity));
            }

            var evaluatedArgs = function.Arguments.Select(Evaluate).ToArray();

            foreach (var ea in evaluatedArgs)
            {
                if (ea.HasError)
                {
                    return Result.Error(ea.ErrorMessage);
                }
            }

            try
            {
                FunctionBody body;

                if (FunctionsLibrary.TryGetValue(function.Definition, out body))
                {
                    return body(this.Context, evaluatedArgs);
                }
                else
                {
                    return Result.Error(string.Format("Funkce {0} nebyla definována.", function.Definition));
                }
            }
            catch (Exception ex)
            {
                return Result.Error(string.Format("Při volání funkce {0} došlo k chybě. {1}", function.Definition, ex.Message));
            }
        }

        private Result EvaluateConstant(Constant constant)
        {
            if (constant.Value is double)
            {
                return Result.Success((double)constant.Value);
            }
            else if (constant.Value is string)
            {
                return Result.Success((string)constant.Value);
            }
            else
            {
                return Result.Success((Element)constant.Value);
            }
        }

        private Exception GetUnknownExpressionException()
        {
            return new ArgumentException("Neznámý druh výrazu.");
        }

        private Result EvaluateUnary(UnaryExpression expression)
        {
            var a = Evaluate(expression.Subexpression);

            if (a.HasError)
            {
                return Result.Error(a.ErrorMessage);
            }

            if (expression is Negation)
            {
                double? cislo = Converters.ToNumberOrNull(a);

                if (cislo != null)
                {
                    return Result.Success(-cislo.Value);
                }
                else
                {
                    return Result.Error("Nečíselnou hodnotu nelze znegovat.");
                }
            }

            throw GetUnknownExpressionException();
        }

        private Result EvaluateSymbol(SymbolExpression expression)
        {
            if (expression.IsElementReference)
            {
                return EvaluateElement(expression.ElementName);
            }
            else
            {
                Element element;

                if (expression.IsRelativePropertyReference)
                {
                    element = Context;
                }
                else
                {
                    var r = EvaluateElement(expression.ElementName);
                    if (r.HasError)
                    {
                        return Result.Error(r.ErrorMessage);
                    }
                    else
                    {
                        element = (Element)r.Value;
                    }
                }

                return EvaluateProperty(element, expression.PropertyNames);
            }
        }

        private Result EvaluateProperty(Element element, IEnumerable<string> propertyNames)
        {
            string name = propertyNames.First();
            Property property;

            if (element.TryGetProperty(name, out property))
            {
                var r = EvaluateProperty(element, property);
                if (r.HasError)
                {
                    return r;
                }
                else
                {
                    // if the value should be element,
                    if (propertyNames.Skip(1).Any())
                    {
                        var valueAsElement = r.Value as Element;
                        if (valueAsElement != null)
                        {
                            return EvaluateProperty(valueAsElement, propertyNames.Skip(1));
                        }
                        else
                        {
                            return Result.Error(string.Format("Value of the {0}.{1} property is not an element.", element, property.Name));
                        }
                    }
                    else
                    {
                        return r;
                    }
                }
            }
            else
            {
                return Result.Error(string.Format("Element {0} nedefinuje vlastnost {1}.", element, name));
            }
        }

        private Result EvaluateProperty(Element element, Property property)
        {
            return Evaluate(element[property.Name].Value, element, SeenExpressions);
        }

        private Result EvaluateElement(string name)
        {
            Element element;

            if (Namespace.Global.TryGetElement(name, out element))
            {
                return Result.Success(element);
            }
            else
            {
                return Result.Error(string.Format("Element \"{0}\" nebyl nalezen.", name));
            }
        }

        private Result EvaluateBinary(BinaryExpression expression)
        {
            var a = Evaluate(expression.Left);
            var b = Evaluate(expression.Right);

            if (a.HasError)
            {
                return Result.Error(a.ErrorMessage);
            }
            if (b.HasError)
            {
                return Result.Error(b.ErrorMessage);
            }

            double? ca = Converters.ToNumberOrNull(a);
            double? cb = Converters.ToNumberOrNull(b);

            if (ca == null)
            {
                return Result.Error(string.Format("Výraz {0} nelze vyhodnotit jako číslo.", expression.Left));
            }
            else if (cb == null)
            {
                return Result.Error(string.Format("Výraz {0} nelze vyhodnotit jako číslo.", expression.Right));
            }
            else
            {
                return ApplyArithmetic(expression, ca.Value, cb.Value);
            }
        }

        private Result ApplyArithmetic(BinaryExpression expression, double a, double b)
        {
            try
            {
                checked
                {
                    double vysledek;

                    if (expression is Addition)
                    {
                        vysledek = a + b;
                    }
                    else if (expression is Substraction)
                    {
                        vysledek = a - b;
                    }
                    else if (expression is Multiplication)
                    {
                        vysledek = a * b;
                    }
                    else if (expression is Division)
                    {
                        if (b != 0)
                        {
                            vysledek = a / b;
                        }
                        else
                        {
                            return Result.Error("Dělení nulou.");
                        }
                    }
                    else if (expression is Modulo)
                    {
                        if (b != 0)
                        {
                            vysledek = a % b;
                        }
                        else
                        {
                            return Result.Error("Zjišťování zbytku po dělení nulou.");
                        }
                    }
                    else
                    {
                        throw GetUnknownExpressionException();
                    }

                    if (double.IsPositiveInfinity(vysledek))
                    {
                        return Result.Error("Číslo je příliš velké.");
                    }
                    else if (double.IsNegativeInfinity(vysledek))
                    {
                        return Result.Error("Číslo je příliš malé.");
                    }
                    else
                    {
                        return Result.Success(vysledek);
                    }
                }
            }
            catch (OverflowException)
            {
                return Result.Error("Číslo je mimo použitelný rozsah.");
            }
            catch (DivideByZeroException)
            {
                return Result.Error("Dělení nulou.");
            }
            catch (ArithmeticException ex)
            {
                return Result.Error(ex.Message);
            }
        }
    }
}
