﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Diagnostics.Contracts;
using System.Diagnostics;

namespace Pikta.Expressions
{
	internal class Lex
	{
        private const RegexOptions SymbolRegexOptions = RegexOptions.CultureInvariant | RegexOptions.IgnoreCase | RegexOptions.Singleline;

		/// <summary>Regulární výraz rozkládající řetězec na tokeny.</summary>
		private static readonly Regex Rx = new Regex(
			@"(?<prazdno>\s+)|" +
			@"(?<cislo>[0-9]+([\.,][0-9]+)?)|" +
			@"(?<operator>[\+\-\*/%])|" +
			@"(?<leva>\()|" +
			@"(" + SymbolRegex + ")|" +
			@"(?<prava>\))|" +
            @"(?<strednik>;)|" +
            @"(?<retezec>\"".*\"")|" +
            @"(?<neplatne>[^\s]+)",
			RegexOptions.CultureInvariant | RegexOptions.IgnoreCase);

		private const string ElementNameRegex = @"(?<element>\p{L}\w*)";
        private const string PropertyNameRegex = @"(?<vlastnost>\p{L}\w*)";
		private const string SymbolRegex = @"(" + ElementNameRegex + @"?(\." + PropertyNameRegex + @")+)|(" + ElementNameRegex + @")";
        private const string StandaloneSymbolRegex = @"\A(" + SymbolRegex + @")\z";
        private const string StandaloneElementNameRegex = @"\A(" + ElementNameRegex + @")\z";
        private const string StandalonePropertyNameRegex = @"\A(" + PropertyNameRegex + @")\z";

		[Pure]
		public static bool IsValidSymbol(string text)
		{
			if(string.IsNullOrWhiteSpace(text))
			{
				return false;
			}
			else
			{
                return Regex.IsMatch(text, StandaloneSymbolRegex, SymbolRegexOptions);
			}
		}

        [Pure]
        public static bool IsValidElementName(string name)
        {
            return !string.IsNullOrWhiteSpace(name) && Regex.IsMatch(name, StandaloneElementNameRegex, SymbolRegexOptions);
        }

        [Pure]
        public static bool IsValidPropertyName(string name)
        {
            return !string.IsNullOrWhiteSpace(name) && Regex.IsMatch(name, StandalonePropertyNameRegex, SymbolRegexOptions);
        }

		/// <summary>Vrátí seznam lexikálních symbolů v zadaném textu.</summary>
		/// <exception cref="Pikta.Expressions.InvalidTokenException">Nalezen neplatný znak.</exception>
		public IList<LexicalSymbol> GetLexicalSymbols(string text)
		{
			Contract.Requires(text != null);

			List<LexicalSymbol> symboly = new List<LexicalSymbol>();

			MatchCollection shody = Rx.Matches(text);

			// projdeme shody
			foreach(Match shoda in shody)
			{
				// pokud je shoda s neplatym symbolem, vyhodime chybu
				if(shoda.Groups["neplatne"].Success)
					throw new InvalidTokenException(string.Format("V zadaném výrazu byl nalezen neplatný řetězec: \"{0}\".",
						shoda.Groups["neplatne"].Value));

				// bile misto preskocime
				if(shoda.Groups["prazdno"].Success)
					continue;

				// zvolime druh tokenu
				LexicalSymbolType druh_vyrazu;
                if (shoda.Groups["cislo"].Success)
                    druh_vyrazu = LexicalSymbolType.Number;
                else if (shoda.Groups["operator"].Success)
                    druh_vyrazu = LexicalSymbolType.Operator;
                else if (shoda.Groups["leva"].Success)
                    druh_vyrazu = LexicalSymbolType.LeftParenthesis;
                else if (shoda.Groups["prava"].Success)
                    druh_vyrazu = LexicalSymbolType.RightParenthesis;
                else if (shoda.Groups["vlastnost"].Success || shoda.Groups["element"].Success)
                    druh_vyrazu = LexicalSymbolType.Symbol;
                else if (shoda.Groups["retezec"].Success)
                    druh_vyrazu = LexicalSymbolType.String;
                else if (shoda.Groups["strednik"].Success)
                    druh_vyrazu = LexicalSymbolType.Semicolon;
                else
                {
                    throw new InvalidTokenException(string.Format("Neznámý token \"{0}\".", shoda.Value));
                }

				// pridame do seznamu
				symboly.Add(new LexicalSymbol(druh_vyrazu, shoda.Value));
			}

			// vratime
			return symboly.ToArray();
		}
	}


	/// <summary>
	/// Token matematického výrazu.
	/// </summary>
	internal class LexicalSymbol
	{
		public LexicalSymbol(LexicalSymbolType type, string value)
		{
			SymbolType = type;
			Value = value;
		}

		/// <summary>Vrátí druh symbolu.</summary>
		public LexicalSymbolType SymbolType { get; private set; }

		/// <summary>Vrátí hodnotu symbolu.</summary>
		public string Value { get; private set; }

		public override string ToString()
		{
			return string.Format("{0}: {1}", SymbolType, Value);
		}
	}

	/// <summary>
	/// Druh lexikálního symbolu.
	/// </summary>
	internal enum LexicalSymbolType
	{
		Number,
		Operator,
		LeftParenthesis,
		RightParenthesis,
		Symbol,
        String,
        Semicolon
	}

	/// <summary>
	/// Nastane při nalezení neplatného znaku.
	/// </summary>
	[Serializable]
	public class InvalidTokenException : Exception
	{
		public InvalidTokenException() { }
		public InvalidTokenException(string message) : base(message) { }
		public InvalidTokenException(string message, Exception inner) : base(message, inner) { }
		protected InvalidTokenException(
		  System.Runtime.Serialization.SerializationInfo info,
		  System.Runtime.Serialization.StreamingContext context)
			: base(info, context) { }
	}
}
