﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Diagnostics.Contracts;
using System.Diagnostics;
using System.Globalization;

namespace Pikta.Expressions
{
	public class Parser
	{
		private static readonly Lex Lex = new Lex();

		private Parser(CultureInfo culture)
		{
			Culture = culture;
		}

		private CultureInfo Culture { get; set; }
		
		/// <summary>Přečte ze zadaného řetězce výraz.</summary>
		/// <exception cref="Pikta.Expressions.InvalidTokenException">Nalezen neplatný znak.</exception>
		/// <exception cref="Pikta.Expressions.ParsingException">Výraz není platný.</exception>
		public static Expression Parse(string text, CultureInfo culture = null)
		{
			var symboly = Lex.GetLexicalSymbols(text);
			// zacneme parsovat
			return new Parser(culture ?? CultureInfo.CurrentCulture).Parse(new ArraySegment<LexicalSymbol>(symboly.ToArray()));
		}

		[Pure]
		public static bool IsValidSymbol(string text)
		{
			return Lex.IsValidSymbol(text);
		}

        [Pure]
        public static bool IsValidElementName(string name) { return Lex.IsValidElementName(name); }

        [Pure]
        public static bool IsValidPropertyName(string name) { return Lex.IsValidPropertyName(name); }

		/// <summary>Převede zadanou posloupnost lexikálních symbolů na výraz.</summary>
		/// <exception cref="Pikta.Expressions.ParsingException">Výraz není platný.</exception>
		private Expression Parse(ArraySegment<LexicalSymbol> subexpr)
		{
			int zac = subexpr.Offset;
			int delka = zac + subexpr.Count;

			// casti vyrazu na aktualni urovni
			List<ExpressionPart> casti = new List<ExpressionPart>();

			// projdeme tokeny
			for(int i = zac; i < delka; i++)
			{
				switch(subexpr.Array[i].SymbolType)
				{
					// leva zavorka
					case LexicalSymbolType.LeftParenthesis:
						// najdeme pro ni pravou zavorku
						int prava = FindRightParenthesis(subexpr, i);
						// pokud jsme nenasli, vyhodime vyjimku
                        if (prava == -1)
                            throw CreateMissingRightParenthesis();
						// provedeme rozklad podvyrazu a pridame si ho
						casti.Add(new ExpressionPart(
							Parse(new ArraySegment<LexicalSymbol>(subexpr.Array, i + 1, prava - i - 1))));
						// preskocime na zavorku
						i = prava;
						break;
					case LexicalSymbolType.Number:
						// preparsujeme cislo
						Expression hodnota;
						if(!TryParseNumber(subexpr.Array[i].Value, out hodnota))
							throw new ParsingException(string.Format("Hodnotu {0} se nepodařilo převést na číslo.",
								subexpr.Array[i].Value));
						// pridame si ho
						casti.Add(new ExpressionPart(hodnota));
						break;
                    case LexicalSymbolType.String:
                        casti.Add(new ExpressionPart(new Constant(subexpr.Array[i].Value.Trim('\"'))));
                        break;
					case LexicalSymbolType.Operator:
						// pridame si ho
						casti.Add(new ExpressionPart(subexpr.Array[i].Value));
						break;
					case LexicalSymbolType.Symbol:
						string[] symboly = subexpr.Array[i].Value.Split('.');
						Contract.Assert(symboly.Length > 0);
                        // simple symbol followed by left parenthesis
                        if (symboly.Length == 1 && i + 1 < delka && subexpr.Array[i + 1].SymbolType == LexicalSymbolType.LeftParenthesis)
                        {
                            // it is function call
                            string functionName = symboly[0];
                            // find the right parenthesis ending the arguments list
                            int right = FindRightParenthesis(subexpr, i + 1);
                            if (right == -1)
                            {
                                throw CreateMissingRightParenthesis();
                            }
                            // parse arguments
                            var functionArguments = ParseArguments(new ArraySegment<LexicalSymbol>(subexpr.Array, i + 1, right - i));
                            // add function call
                            casti.Add(new ExpressionPart(
                                new FunctionCall(
                                    new FunctionDefinition(functionName, functionArguments.Count),
                                    functionArguments)));
                            // skip to the last parenthesis
                            i = right;
                        }
                        else
                        {
                            // otherwise, it is just symbol
                            casti.Add(new ExpressionPart(new SymbolExpression(symboly[0], symboly.Skip(1).ToArray())));
                        }
						break;
					
                    case LexicalSymbolType.RightParenthesis:
						throw new ParsingException("Nalezena samostatná pravá závorka.");
                    
                    case LexicalSymbolType.Semicolon:
                        throw new ParsingException("Středník byl použit jinak než jako oddělovač parametrů ve volání funkce.");
					
                    default:
						throw new ArgumentException("Neznámý druh tokenu " + subexpr.Array[i].SymbolType);
				}
			}

            if (casti.Count == 0)
            {
                throw new ParsingException("Byl nalezen prázdný výraz.");
            }

			// z casti na aktualni urovni udelame vyraz
			return MakeExpression(casti);
		}

        private ParsingException CreateMissingRightParenthesis()
        {
            return new ParsingException("Byla nalezena neukončená závorka.");
        }

        private IList<Expression> ParseArguments(ArraySegment<LexicalSymbol> arguments)
        {
            Contract.Requires(arguments.Count >= 2);

            var functionArgs = new List<Expression>();

            if (arguments.Count > 2)
            {
                // take arguments, skip enclosing parentheses
                var argumentsPart = arguments.Array.Skip(arguments.Offset + 1).Take(arguments.Count - 2);
                //var allSemicolons = argumentsPart
                //                    .Select((s, i) => new { Symbol = s, Index = i })
                //                    .Where(p => p.Symbol.SymbolType == LexicalSymbolType.Semicolon)
                //                    .Select(p => p.Index - arguments.Offset);

                // current level semicolons indices relative to arguments.Offset+1
                var semicolonIndices = GetCurrentLevelSemicolons(argumentsPart.ToArray());

                int argStartIndex = arguments.Offset + 1;

                for (int i = 0; i <= semicolonIndices.Count; i++)
                {
                    int argLength;

                    if (i < semicolonIndices.Count)
                    {
                        argLength = arguments.Offset + 1 + semicolonIndices[i] - argStartIndex;
                    }
                    else
                    {
                        argLength = (arguments.Offset + arguments.Count - 1) - argStartIndex;
                    }

                    functionArgs.Add(Parse(new ArraySegment<LexicalSymbol>(arguments.Array, argStartIndex, argLength)));

                    argStartIndex += argLength + 1;
                }
            }

            return functionArgs;
        }

        private IList<int> GetCurrentLevelSemicolons(IList<LexicalSymbol> symbols)
        {
            var indices = new List<int>();

            for (int i = 0; i < symbols.Count; i++)
            {
                switch (symbols[i].SymbolType)
                {
                    case LexicalSymbolType.Semicolon:
                        indices.Add(i);
                        break;
                    case LexicalSymbolType.LeftParenthesis:
                        // skip parenthesis block
                        var right = FindRightParenthesis(new ArraySegment<LexicalSymbol>(symbols.ToArray()), i);
                        if (right == -1)
                        {
                            throw CreateMissingRightParenthesis();
                        }
                        i = right;
                        break;
                    default:
                        break;
                }
            }

            return indices;
        }

		/// <summary>
		/// Převede zadané číslo. 
		/// </summary>
		private bool TryParseNumber(string str, out Expression result)
		{
			double cislo;

			if(double.TryParse(str, NumberStyles.Float, Culture, out cislo))
			{
				result = new Constant(cislo);
				return true;
			}

			result = null;
			return false;
		}

		/// <summary>
		/// Najde k zadané levé závorce odpovídající pravou. Nepodaří-li se, vrátí -1.
		/// </summary>
		private static int FindRightParenthesis(ArraySegment<LexicalSymbol> expression, int leftParenthesisIndex)
		{
			int uroven = 1;

			// projdeme tokeny podvyrazu
			for(int i = leftParenthesisIndex + 1; i < expression.Offset + expression.Count; i++)
			{
				switch(expression.Array[i].SymbolType)
				{
					case LexicalSymbolType.LeftParenthesis:
						uroven++;
						break;
					case LexicalSymbolType.RightParenthesis:
						uroven--;
						// pokud je na stejne urovni, nasli jsme ji
						if(uroven == 0)
							return i;
						break;
				}
			}

			// nenasli jsme
			return -1;
		}

		/// <summary>
		/// Část výrazu. Buď podvýraz, nebo operátor.
		/// </summary>
		private class ExpressionPart
		{
			public ExpressionPart(Expression expression)
			{
				if(expression == null) throw new ArgumentNullException("expression");
				Expression = expression;
			}

			public ExpressionPart(string op)
			{
				if(op == null) throw new ArgumentNullException("op");
				Operator = op;
			}

			public Expression Expression { get; private set; }
			public string Operator { get; private set; }

			public override string ToString()
			{
				if(Expression != null)
					return Expression.ToString();
				else
					return Operator;
			}
		}

		/// <summary>
		/// Podle priority operátorů složí z daných částí výraz.
		/// </summary>
		private static Expression MakeExpression(IList<ExpressionPart> parts)
		{
			Contract.Requires(parts.Count > 0, "Seznam částí nesmí být prázdný.");

			int nejvyssi;

			// vezmeme dalsi op s nejvyssi prioritou
			while((nejvyssi = FindOperatorWithHighiestPriority(parts)) >= 0)
			{
				string op = parts[nejvyssi].Operator;
				bool muze_byt_unarni = op == "-";

				// zjistime indexy operandu
				int levy = nejvyssi - 1;
				int pravy = nejvyssi + 1;

				// pokud nema nektery operand, je to zle
				if(!muze_byt_unarni && levy < 0) throw new ParsingException(string.Format("Operátor {0} nemá svůj levý operand.", op));
				if(pravy == parts.Count) throw new ParsingException(string.Format("Operátor {0} nemá svůj pravý operand.", op));

				Expression l_op = levy >= 0 ? parts[levy].Expression : null;
				Expression p_op = parts[pravy].Expression;

				// pokud operandem neni vyraz, je to taky spatne
				if(muze_byt_unarni)
				{
					if(p_op == null)
						throw new ParsingException(string.Format("Operátor {0} musí mít jako operand jeden, nebo dva výrazy.", op));
				}
				else
				{
					if(l_op == null || p_op == null)
						throw new ParsingException(string.Format("Operátor {0} musí mít jako operandy dva výrazy.", op));
				}

				int pocet_operandu = l_op != null ? 2 : 1;

				// smazeme ze seznamu operator s jeho operandy
				int pozice = l_op != null ? levy : pravy - 1;
				for(int j = 0; j < pocet_operandu + 1; j++)
					parts.RemoveAt(pozice);

				// vlozime misto nich novy vyraz
				if(pocet_operandu == 2)
					parts.Insert(pozice, new ExpressionPart(CreateBinaryExpression(op, l_op, p_op)));
				else
					parts.Insert(pozice, new ExpressionPart(CreateUnaryExpression(op, p_op)));
			}

			// pokud uz mame jen jednu cast,
			if(parts.Count == 1)
			{
				// je to hledany vyraz
				Contract.Assert(parts[0].Expression != null, "Nezbyl nám výraz.. podivné.");
				return parts[0].Expression;
			}
			// jinak tam zrejme chybi operator
			else
			{
				throw new ParsingException("V podvýrazu není uveden operátor.");
			}
		}

		/// <summary>
		/// Vytvoří unární výraz podle zadaného operátoru.
		/// </summary>
		private static Expression CreateUnaryExpression(string op, Expression operand)
		{
			switch(op)
			{
				case "-":
					return new Negation(operand);
				default:
					Debug.Fail("Neznámý operátor " + op);
					throw new NotImplementedException();
			}
		}

		/// <summary>
		/// Vytvoří binární výraz podle zadaného operátoru.
		/// </summary>
		private static BinaryExpression CreateBinaryExpression(string op, Expression left, Expression right)
		{
			switch(op)
			{
				case "+": return new Addition(left, right);
				case "-": return new Substraction(left, right);
				case "*": return new Multiplication(left, right);
				case "/": return new Division(left, right);
				case "%": return new Modulo(left, right);
				default:
					Debug.Fail("Neznámý operátor " + op);
					throw new NotImplementedException();
			}
		}

		private static readonly Dictionary<string, int> Priorities = new Dictionary<string, int>() { { "*", 1 }, { "/", 1 }, { "%", 1}, { "+", 2 }, { "-", 2 } };

		/// <summary>
		/// Najde v zadaném seznamu nejlevější operátor s nejvyšší prioritou. Pokud seznam žádný
		/// operátor neobsahuje, vrátí -1.
		/// </summary>
		private static int FindOperatorWithHighiestPriority(IList<ExpressionPart> parts)
		{
			int min = int.MaxValue;
			int op_index = -1;

			// projdeme casti
			for(int i = 0; i < parts.Count; i++)
			{
				// je-li to operator,
				string op = parts[i].Operator;
				if(op != null)
				{
					// hledame min
					int priorita = Priorities[op];
					if(priorita < min)
					{
						min = priorita;
						op_index = i;
					}
				}
			}

			Debug.Assert(op_index == -1 || parts[op_index].Operator != null, "Nejvyšší prioritu mělo něco jiného než operátor...");

			return op_index;
		}
	}

	/// <summary>
	/// Nastane při chybě během parsování výrazu.
	/// </summary>
	[Serializable]
	public class ParsingException : Exception
	{
		public ParsingException() { }
		public ParsingException(string message) : base(message) { }
		public ParsingException(string message, Exception inner) : base(message, inner) { }
		protected ParsingException(
		  System.Runtime.Serialization.SerializationInfo info,
		  System.Runtime.Serialization.StreamingContext context)
			: base(info, context) { }
	}
}
