﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics.Contracts;
using System.Text.RegularExpressions;
using System.Globalization;

namespace Pikta.Expressions
{
    // TODO: Co takhle udelat tridu Symbol misto predavani stringu?

    /// <summary>
    /// Symbolický výraz; představuje buď odkaz na element, nebo na jeho vlastnost.
    /// </summary>
	public class SymbolExpression : Expression
	{
		public SymbolExpression(string elementName, params string[] propertyNames)
		{
			Contract.Requires(!string.IsNullOrEmpty(elementName) || propertyNames.Length > 0);
			Contract.Requires(Contract.ForAll(propertyNames, name => !string.IsNullOrEmpty(name)));

			if(elementName == string.Empty)
				elementName = null;

			ElementName = elementName;
			PropertyNames = propertyNames;
		}

		/// <summary>
		/// Vrátí název odkazovaného elementu. Může být null, pokud se jedná o relativní odkaz.
		/// </summary>
		public string ElementName { get; private set; }
		
		/// <summary>
		/// Vrátí seznam názvů odkazovaných vlastností. Může být prázdný, pokud se jedná o odkaz na element.
		/// </summary>
		public IList<string> PropertyNames { get; private set; }

		public bool IsElementReference { get { return ElementName != null && PropertyNames.Count == 0; } }

		public bool IsPropertyReference { get { return PropertyNames.Count > 0; } }

		public bool IsRelativePropertyReference { get { return IsPropertyReference && ElementName == null; } }

		public override IEnumerable<SymbolExpression> AllSymbolExpressions { get { yield return this; } }

        public override IEnumerable<Constant> AllConstants
        {
            get { return Enumerable.Empty<Constant>(); }
        }

		public override string ToString()
		{
            return ToString(CultureInfo.CurrentCulture);
		}

        public override string ToString(IFormatProvider formatProvider)
        {
            if (IsElementReference)
            {
                return ElementName;
            }
            else
            {
                return string.Format(formatProvider, "{0}.{1}", ElementName, string.Join(".", PropertyNames));
            }
        }

		public bool Equals(SymbolExpression other)
		{
			return other != null && NameComparer.Instance.Equals(ElementName, other.ElementName) && PropertyNames.Count == other.PropertyNames.Count &&
				PropertyNames.SequenceEqual(other.PropertyNames, NameComparer.Instance);
		}

		public override bool Equals(Expression other)
		{
			return Equals(other as SymbolExpression);
		}

		public override bool Equals(object obj)
		{
			return Equals(obj as SymbolExpression);
		}

		public override int GetHashCode()
		{
            int propertyHash = 0;

            if (PropertyNames.Any())
            {
                propertyHash = PropertyNames.Select(str => str.GetHashCode()).Aggregate((a, b) => a ^ b);
            }

			return (ElementName != null ? ElementName.GetHashCode() : 0) ^ propertyHash;
		}
	}
}
