﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Contracts;
using System.Linq;
using System.Text;
using Pikta.Expressions;

namespace Pikta
{
	public class Property : IEquatable<Property>
	{
		public Property(Element element, string name)
		{
			Element = element;
			Name = name;
		}

		public Element Element { get; private set; }
		
		public string Name { get; private set; }

        /// <summary>
        /// Gets the property definition. If the property is not defined, returns null.
        /// </summary>
		public PropertyDefinition Definition
		{
			get { return Element.GetPropertyDefinition(Name); }
		}

        /// <summary>
        /// Gets a value indicating whether the property is localy defined in the <see cref="Element"/>.
        /// </summary>
        public bool IsLocal
        {
            get { return Definition != null && Definition.Element == Element; }
        }

		public Expression Value
		{
			get
			{
				if(Definition != null)
				{
					return Definition.Value;
				}
				else
				{
					throw new InvalidOperationException(string.Format("Vlastnost {0} není v elementu {1} definována.", Name, Element));
				}
			}
			set
			{
                SetValue(value, InferType(value));
			}
		}

        public void SetValue(Expression value, PropertyType type = null)
        {
            if (IsLocal)
            {
                Definition.Value = value;
            }
            else
            {
                Element.LocalPropertyDefinitions.Add(new PropertyDefinition(Element, Name, type) { Value = value });
            }
        }

        public override bool Equals(object obj)
        {
            return Equals(obj as Property);
        }

        public override int GetHashCode()
        {
            return Element.GetHashCode() ^ NameComparer.Instance.GetHashCode(Name);
        }

        public bool Equals(Property other)
        {
            return other != null && Element.Equals(other.Element) && NameComparer.Instance.Equals(Name, other.Name);
        }

        public override string ToString()
        {
            return string.Format("{0}.{1}", Element, Name);
        }

        /// <summary>
        /// Removes local property definition, if there is any.
        /// </summary>
        /// <exception cref="System.InvalidOperationException">The property is not defined.</exception>
        public void Remove()
        {
            var def = Definition;

            if (def == null)
            {
                throw new InvalidOperationException(string.Format("Vlastnost \"{0}\" není definována. Nelze ji odebrat.", Name));
            }

            def.Element.LocalPropertyDefinitions.Remove(def);
        }

        /// <summary>
        /// Odvodí typ hodnoty, která je výsledkem vyhodnocení zadaného výrazu.
        /// </summary>
        /// <exception cref="System.ArgumentException">Typ se nepodařilo odvodit.</exception>
        private PropertyType InferType(Expression expression)
        {
            Contract.Requires(expression != null);

            Result vysledek = Element.Evaluate(expression);

            if (vysledek.HasError)
            {
                return PropertyType.Any;
                //throw new ArgumentException(string.Format("Typ hodnoty výrazu \"{0}\" nelze určit, protože se ho nepodařilo vyhodnotit. Chyba vyhodnocování: {1}", expression, vysledek.ErrorMessage));
            }

            return vysledek.ValueType;
        }
    }
}
